from enum import StrEnum
from typing import Any, Literal

from pydantic import BaseModel, model_validator


class SettingsFieldType(StrEnum):
    TEXT = "text"
    NUMBER = "number"
    SELECT = "select"
    CHECKBOX = "checkbox"
    TEXTAREA = "textarea"
    PASSWORD = "password"
    EMAIL = "email"
    DATE = "date"
    IMAGE = "image"


class SettingsSelectOption(BaseModel):
    value: str
    label: str


class SettingsField(BaseModel):
    id: str
    type: SettingsFieldType
    label: str
    required: bool = False
    readonly: bool = False
    value: Any = None
    is_internal: bool = False

    placeholder: str | None = None
    options: list[SettingsSelectOption] | None = None
    min: int | float | None = None
    max: int | float | None = None
    rows: int | None = None
    src: str | None = None
    width: int | None = None
    height: int | None = None


class SettingsFieldGroup(BaseModel):
    id: str
    label: str
    fields: list[SettingsField]
    collapsible: Literal["open", "closed"] | None = None
    is_internal: bool = False


def _collect_all_fields(items: list["SettingsField | SettingsFieldGroup"]) -> list[SettingsField]:
    result: list[SettingsField] = []
    for item in items:
        if isinstance(item, SettingsFieldGroup):
            result.extend(item.fields)
        else:
            result.append(item)
    return result


class SettingsSchema(BaseModel):
    version: str = "1.0"
    fields: list[SettingsField | SettingsFieldGroup]

    @model_validator(mode="after")
    def validate_unique_ids(self) -> "SettingsSchema":
        all_fields = _collect_all_fields(self.fields)
        ids = [f.id for f in all_fields]
        if len(ids) != len(set(ids)):
            raise ValueError("Field IDs must be unique")
        return self

    def get_field(self, field_id: str) -> SettingsField | None:
        for item in self.fields:
            if isinstance(item, SettingsFieldGroup):
                for field in item.fields:
                    if field.id == field_id:
                        return field
            elif item.id == field_id:
                return item
        return None

    def get_defaults(self) -> dict[str, Any]:
        return {f.id: f.value for f in _collect_all_fields(self.fields)}

    def get_public_fields(self) -> list["SettingsField | SettingsFieldGroup"]:
        result: list[SettingsField | SettingsFieldGroup] = []
        for item in self.fields:
            if isinstance(item, SettingsFieldGroup):
                if item.is_internal:
                    continue
                public_children = [f for f in item.fields if not f.is_internal]
                if public_children:
                    group_copy = item.model_copy(update={"fields": public_children})
                    result.append(group_copy)
            elif not item.is_internal:
                result.append(item)
        return result



from __future__ import annotations

import uuid
from datetime import UTC, datetime
from typing import TYPE_CHECKING

from dooers.protocol.frames import (
    EventAppendPayload,
    S2C_EventAppend,
    S2C_ThreadUpsert,
    ThreadUpsertPayload,
)
from dooers.protocol.models import Actor, ContentPart, Thread, ThreadEvent, User
from dooers.protocol.parser import serialize_frame

if TYPE_CHECKING:
    from dooers.persistence.base import Persistence
    from dooers.registry import ConnectionRegistry


def _generate_id() -> str:
    return str(uuid.uuid4())


def _now() -> datetime:
    return datetime.now(UTC)


class BroadcastManager:
    def __init__(
        self,
        registry: ConnectionRegistry,
        persistence: Persistence,
        subscriptions: dict[str, set[str]],
    ) -> None:
        self._registry = registry
        self._persistence = persistence
        self._subscriptions = subscriptions

    async def send_event(
        self,
        worker_id: str,
        thread_id: str,
        content: list[ContentPart],
        actor: Actor = "system",
        user: User | None = None,
    ) -> tuple[ThreadEvent, int]:
        thread = await self._persistence.get_thread(thread_id)
        if not thread:
            raise ValueError(f"Thread {thread_id} not found")
        if thread.worker_id != worker_id:
            raise PermissionError(f"Thread {thread_id} belongs to different worker")

        now = _now()

        event = ThreadEvent(
            id=_generate_id(),
            thread_id=thread_id,
            run_id=None,
            type="message",
            actor=actor,
            user=user or User(user_id=""),
            content=content,
            data=None,
            created_at=now,
        )

        await self._persistence.create_event(event)

        thread.last_event_at = now
        thread.updated_at = now
        await self._persistence.update_thread(thread)

        frame = S2C_EventAppend(
            id=_generate_id(),
            payload=EventAppendPayload(thread_id=thread_id, events=[event]),
        )
        message = serialize_frame(frame)
        count = await self._registry.broadcast(worker_id, message)

        return event, count

    async def send_thread_update(
        self,
        worker_id: str,
        thread: Thread,
    ) -> int:
        if thread.worker_id != worker_id:
            raise PermissionError(f"Thread {thread.id} belongs to different worker")

        frame = S2C_ThreadUpsert(
            id=_generate_id(),
            payload=ThreadUpsertPayload(thread=thread),
        )
        message = serialize_frame(frame)
        return await self._registry.broadcast(worker_id, message)

    async def create_thread_and_broadcast(
        self,
        worker_id: str,
        user: User | None = None,
        organization_id: str = "",
        workspace_id: str = "",
        title: str | None = None,
    ) -> tuple[Thread, int]:
        now = _now()
        owner = user or User(user_id="")
        thread = Thread(
            id=_generate_id(),
            worker_id=worker_id,
            organization_id=organization_id,
            workspace_id=workspace_id,
            owner=owner,
            users=[owner] if owner.user_id else [],
            title=title,
            created_at=now,
            updated_at=now,
            last_event_at=now,
        )

        await self._persistence.create_thread(thread)

        frame = S2C_ThreadUpsert(
            id=_generate_id(),
            payload=ThreadUpsertPayload(thread=thread),
        )
        message = serialize_frame(frame)
        count = await self._registry.broadcast(worker_id, message)

        return thread, count



from __future__ import annotations

from collections.abc import AsyncGenerator
from typing import TYPE_CHECKING

from dooers.handlers.send import WorkerEvent

if TYPE_CHECKING:
    from dooers.handlers.pipeline import HandlerContext, HandlerPipeline, PipelineResult


class DispatchStream:
    """Async-iterable stream of handler events from a programmatic dispatch.

    Properties ``thread_id``, ``event_id``, and ``is_new_thread`` are
    available immediately (before iteration), because the pipeline setup
    phase runs before this object is returned.
    """

    def __init__(
        self,
        pipeline: HandlerPipeline,
        context: HandlerContext,
        result: PipelineResult,
    ) -> None:
        self._pipeline = pipeline
        self._context = context
        self._result = result

    @property
    def thread_id(self) -> str:
        return self._result.thread.id

    @property
    def event_id(self) -> str:
        return self._result.user_event.id

    @property
    def is_new_thread(self) -> bool:
        return self._result.is_new_thread

    def __aiter__(self) -> AsyncGenerator[WorkerEvent, None]:
        return self._pipeline.execute(self._context, self._result)

    async def collect(self) -> list[WorkerEvent]:
        events: list[WorkerEvent] = []
        async for event in self:
            events.append(event)
        return events



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Literal, Protocol

from dooers.protocol.models import Run, Thread, ThreadEvent, User

if TYPE_CHECKING:
    from dooers.features.analytics.models import AnalyticsEventPayload

EventOrder = Literal["asc", "desc"]

FILTERABLE_FIELDS = {"type", "actor", "user_id", "user_name", "user_email", "run_id"}


class Persistence(Protocol):
    async def connect(self) -> None: ...
    async def disconnect(self) -> None: ...
    async def migrate(self) -> None: ...
    async def create_thread(self, thread: Thread) -> None: ...
    async def get_thread(self, thread_id: str) -> Thread | None: ...
    async def update_thread(self, thread: Thread) -> None: ...
    async def list_threads(
        self,
        worker_id: str,
        organization_id: str,
        workspace_id: str,
        user_id: str | None,
        cursor: str | None,
        limit: int,
        scope: str = "member",
        user_email: str | None = None,
        identity_ids: list[str] | None = None,
    ) -> list[Thread]: ...
    async def count_threads(
        self,
        worker_id: str,
        organization_id: str,
        workspace_id: str,
        user_id: str | None,
        scope: str = "member",
        user_email: str | None = None,
        identity_ids: list[str] | None = None,
    ) -> int: ...
    async def delete_thread(self, thread_id: str) -> None: ...
    async def create_event(self, event: ThreadEvent) -> None: ...
    async def get_events(
        self,
        thread_id: str,
        *,
        after_event_id: str | None = None,
        before_event_id: str | None = None,
        limit: int = 50,
        order: EventOrder = "asc",
        filters: dict[str, str] | None = None,
    ) -> list[ThreadEvent]: ...
    async def create_run(self, run: Run) -> None: ...
    async def update_run(self, run: Run) -> None: ...
    async def get_event(self, event_id: str) -> ThreadEvent | None: ...
    async def update_event(self, event: ThreadEvent) -> None: ...
    async def delete_event(self, event_id: str) -> None: ...
    async def get_run(self, run_id: str) -> Run | None: ...
    async def list_runs(
        self,
        thread_id: str | None = None,
        worker_id: str | None = None,
        status: str | None = None,
        limit: int = 50,
    ) -> list[Run]: ...
    async def upsert_thread_participant(self, thread_id: str, user: User) -> None: ...

    async def get_settings(self, worker_id: str) -> dict[str, Any]: ...
    async def update_setting(self, worker_id: str, field_id: str, value: Any) -> datetime: ...
    async def set_settings(self, worker_id: str, values: dict[str, Any]) -> datetime: ...

    async def insert_analytics_events(self, events: list[AnalyticsEventPayload]) -> None: ...



from typing import TYPE_CHECKING, Any, Literal

from dooers.protocol.models import ThreadEvent

if TYPE_CHECKING:
    from dooers.persistence.base import EventOrder, Persistence

HistoryFormat = Literal["openai", "anthropic", "google", "cohere", "voyage"]


class WorkerMemory:
    def __init__(self, thread_id: str, persistence: "Persistence"):
        self._thread_id = thread_id
        self._persistence = persistence

    async def get_history_raw(
        self,
        limit: int = 50,
        order: "EventOrder" = "asc",
        filters: dict[str, str] | None = None,
    ) -> list[ThreadEvent]:
        return await self._persistence.get_events(
            self._thread_id,
            limit=limit,
            order=order,
            filters=filters,
        )

    async def get_history(
        self,
        limit: int = 50,
        format: HistoryFormat = "openai",
        order: "EventOrder" = "desc",
        filters: dict[str, str] | None = None,
    ) -> list[dict[str, Any]]:
        raw_limit = limit * 3
        events = await self.get_history_raw(raw_limit, order=order, filters=filters)

        if order == "desc":
            events.reverse()

        messages: list[dict[str, Any]] = []
        for event in events:
            if event.type != "message" or not event.content:
                continue

            text = self._extract_text_with_media(event.content)
            if not text:
                continue

            messages.append(self._format_message(text, event.actor == "user", format))

        if order == "desc":
            return messages[-limit:]
        return messages[:limit]

    @staticmethod
    def _extract_text_with_media(content: list) -> str:
        """Extract text from content parts, adding indicators for media parts."""
        segments: list[str] = []
        for part in content:
            # Handle both Pydantic models and raw dicts (from DB deserialization)
            if isinstance(part, dict):
                p_type = part.get("type")
                if p_type == "text" and part.get("text"):
                    segments.append(part["text"])
                elif p_type == "audio":
                    fname = part.get("filename") or "audio"
                    mime = part.get("mime_type") or ""
                    dur = part.get("duration")
                    dur_str = f", {dur}s" if dur else ""
                    segments.append(f"[audio: {fname}{', ' + mime if mime else ''}{dur_str}]")
                elif p_type == "image":
                    fname = part.get("filename") or "image"
                    segments.append(f"[image: {fname}]")
                elif p_type == "document":
                    fname = part.get("filename") or "document"
                    segments.append(f"[document: {fname}]")
            else:
                if hasattr(part, "text") and part.text:
                    segments.append(part.text)
                elif hasattr(part, "type"):
                    if part.type == "audio":
                        fname = getattr(part, "filename", None) or "audio"
                        mime = getattr(part, "mime_type", None) or ""
                        dur = getattr(part, "duration", None)
                        dur_str = f", {dur}s" if dur else ""
                        segments.append(f"[audio: {fname}{', ' + mime if mime else ''}{dur_str}]")
                    elif part.type == "image":
                        fname = getattr(part, "filename", None) or "image"
                        segments.append(f"[image: {fname}]")
                    elif part.type == "document":
                        fname = getattr(part, "filename", None) or "document"
                        segments.append(f"[document: {fname}]")
        return " ".join(segments)

    def _format_message(
        self,
        text: str,
        is_user: bool,
        format: HistoryFormat,
    ) -> dict[str, Any]:
        match format:
            case "openai" | "voyage":
                return {
                    "role": "user" if is_user else "assistant",
                    "content": text,
                }
            case "anthropic":
                return {
                    "role": "user" if is_user else "assistant",
                    "content": text,
                }
            case "google":
                return {
                    "role": "user" if is_user else "model",
                    "parts": [{"text": text}],
                }
            case "cohere":
                return {
                    "role": "USER" if is_user else "CHATBOT",
                    "message": text,
                }
            case _:
                return {
                    "role": "user" if is_user else "assistant",
                    "content": text,
                }



"""Multiple entry points: WebSocket for real-time chat, REST for dispatch.

- WebSocket /ws         — real-time chat via handler
- POST /api/summarize   — dispatch a summarization handler on an existing thread
- POST /api/notify      — inject a system notification into a thread via dispatch
"""

from fastapi import FastAPI, Request, WebSocket
from fastapi.responses import JSONResponse

from dooers import WorkerConfig, WorkerServer

app = FastAPI()
worker_server = WorkerServer(
    WorkerConfig(
        database_type="sqlite",
        database_name="worker.db",
        assistant_name="Multi Bot",
    )
)

WORKER_ID = "multi-worker"
ORG_ID = "org_demo"
WS_ID = "ws_demo"


# --- Handlers (identical API for WebSocket and dispatch) ---


async def chat_handler(incoming, send, memory, analytics, settings):
    yield send.run_start(agent_id="chat")
    yield send.text(f"You said: {incoming.message}")
    yield send.update_thread(title=incoming.message[:60])
    yield send.run_end()


async def summarize_handler(incoming, send, memory, analytics, settings):
    yield send.run_start(agent_id="summarizer")

    history = await memory.get_history(limit=50)
    message_count = len(history)

    summary = f"Thread summary: {message_count} messages. Latest: {incoming.message[:80]}"

    yield send.text(summary)
    yield send.update_thread(title="Summary")
    yield send.run_end()


async def notify_handler(incoming, send, memory, analytics, settings):
    yield send.run_start(agent_id="notifier")
    yield send.text(incoming.message, author="System")
    yield send.run_end()


# --- WebSocket entry point (real-time) ---


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    await worker_server.handle(websocket, chat_handler)


# --- REST entry points (dispatch) ---


@app.post("/api/summarize")
async def summarize(request: Request):
    body = await request.json()
    thread_id = body.get("thread_id")
    if not thread_id:
        return JSONResponse({"error": "thread_id required"}, status_code=400)

    stream = await worker_server.dispatch(
        handler=summarize_handler,
        worker_id=WORKER_ID,
        organization_id=ORG_ID,
        workspace_id=WS_ID,
        message="Summarize this conversation",
        user_id="api",
        thread_id=thread_id,
    )

    texts = []
    async for event in stream:
        if event.send_type == "text":
            texts.append(event.data["text"])

    return {"thread_id": stream.thread_id, "summary": " ".join(texts)}


@app.post("/api/notify")
async def notify(request: Request):
    body = await request.json()
    thread_id = body.get("thread_id")
    message = body.get("message", "")
    if not thread_id or not message:
        return JSONResponse({"error": "thread_id and message required"}, status_code=400)

    stream = await worker_server.dispatch(
        handler=notify_handler,
        worker_id=WORKER_ID,
        organization_id=ORG_ID,
        workspace_id=WS_ID,
        message=message,
        user_id="system",
        thread_id=thread_id,
    )

    await stream.collect()

    return {"status": "ok", "thread_id": stream.thread_id}


@app.on_event("startup")
async def startup():
    await worker_server.ensure_initialized()


@app.on_event("shutdown")
async def shutdown():
    await worker_server.close()



import os

from fastapi import FastAPI, WebSocket
from langchain_core.messages import AIMessage, HumanMessage, SystemMessage
from langchain_core.tools import tool
from langchain_openai import ChatOpenAI
from langgraph.prebuilt import create_react_agent

from dooers import WorkerConfig, WorkerServer

app = FastAPI()
worker_server = WorkerServer(
    WorkerConfig(
        database_type="sqlite",
        database_name="worker.db",
    )
)

llm = ChatOpenAI(
    model="gpt-4o-mini",
    api_key=os.getenv("OPENAI_API_KEY"),
)


@tool
def multiply(a: int, b: int) -> int:
    return a * b


@tool
def add(a: int, b: int) -> int:
    return a + b


agent = create_react_agent(llm, tools=[multiply, add])


async def langgraph_agent(incoming, send, memory, analytics, settings):
    yield send.run_start(agent_id="langgraph")

    # get_history_raw() returns ThreadEvent objects for manual conversion
    events = await memory.get_history_raw(limit=20, order="asc")

    messages = [SystemMessage(content="You are a helpful assistant with math tools.")]
    for event in events:
        if event.type == "message" and event.content:
            text = " ".join(part.text for part in event.content if hasattr(part, "text"))
            if text:
                if event.actor == "user":
                    messages.append(HumanMessage(content=text))
                else:
                    messages.append(AIMessage(content=text))

    messages.append(HumanMessage(content=incoming.message))

    result = await agent.ainvoke({"messages": messages})

    for msg in result["messages"]:
        if hasattr(msg, "tool_calls") and msg.tool_calls:
            for tc in msg.tool_calls:
                yield send.tool_call(tc["name"], tc["args"])
        if msg.type == "tool":
            yield send.tool_result(msg.name, {"result": msg.content})

    final = result["messages"][-1]
    if final.content:
        yield send.text(final.content)

    yield send.update_thread(title=incoming.message[:60])
    yield send.run_end()


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    await worker_server.handle(websocket, langgraph_agent)


@app.on_event("shutdown")
async def shutdown():
    await worker_server.close()



import os

from anthropic import AsyncAnthropic
from fastapi import FastAPI, WebSocket

from dooers import (
    SettingsField,
    SettingsFieldType,
    SettingsSchema,
    SettingsSelectOption,
    WorkerConfig,
    WorkerServer,
)

app = FastAPI()

settings_schema = SettingsSchema(
    fields=[
        SettingsField(
            id="model",
            type=SettingsFieldType.SELECT,
            label="Model",
            value="claude-sonnet-4-20250514",
            options=[
                SettingsSelectOption(value="claude-sonnet-4-20250514", label="Claude Sonnet 4"),
                SettingsSelectOption(value="claude-opus-4-20250514", label="Claude Opus 4"),
                SettingsSelectOption(value="claude-haiku-4-20250514", label="Claude Haiku 4"),
            ],
        ),
        SettingsField(
            id="max_tokens",
            type=SettingsFieldType.NUMBER,
            label="Max Tokens",
            value=1024,
            min=1,
            max=4096,
        ),
        SettingsField(
            id="system_prompt",
            type=SettingsFieldType.TEXTAREA,
            label="System Prompt",
            value="You are a helpful assistant.",
            rows=4,
        ),
    ]
)

worker_server = WorkerServer(
    WorkerConfig(
        database_type="sqlite",
        database_name="worker.db",
        settings_schema=settings_schema,
    )
)
client = AsyncAnthropic(api_key=os.getenv("ANTHROPIC_API_KEY"))


async def anthropic_agent(incoming, send, memory, analytics, settings):
    yield send.run_start(agent_id="anthropic-claude")

    model = await settings.get("model")
    max_tokens = await settings.get("max_tokens")
    system_prompt = await settings.get("system_prompt")

    await analytics.track("llm.request", data={"model": model})

    # get_history() returns messages formatted for the LLM provider
    history = await memory.get_history(limit=20, format="anthropic")

    messages = list(history)
    messages.append({"role": "user", "content": incoming.message})

    result = await client.messages.create(
        model=model,
        max_tokens=int(max_tokens),
        system=system_prompt,
        messages=messages,
    )

    await analytics.track(
        "llm.response",
        data={
            "model": model,
            "input_tokens": result.usage.input_tokens,
            "output_tokens": result.usage.output_tokens,
        },
    )

    yield send.text(result.content[0].text)
    yield send.update_thread(title=incoming.message[:60])
    yield send.run_end()


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    await worker_server.handle(websocket, anthropic_agent)


@app.on_event("shutdown")
async def shutdown():
    await worker_server.close()



from .broadcaster import SettingsBroadcaster
from .models import SettingsField, SettingsFieldGroup, SettingsFieldType, SettingsSchema, SettingsSelectOption
from .worker_settings import WorkerSettings

__all__ = [
    "SettingsBroadcaster",
    "SettingsField",
    "SettingsFieldGroup",
    "SettingsFieldType",
    "SettingsSchema",
    "SettingsSelectOption",
    "WorkerSettings",
]



from .collector import AnalyticsCollector
from .models import AnalyticsBatch, AnalyticsEvent, AnalyticsEventPayload
from .worker_analytics import WorkerAnalytics

__all__ = [
    "AnalyticsCollector",
    "AnalyticsBatch",
    "AnalyticsEvent",
    "AnalyticsEventPayload",
    "WorkerAnalytics",
]



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    pass


class WebSocketProtocol(Protocol):
    async def send_text(self, data: str) -> None: ...
    async def receive_text(self) -> str: ...


class ConnectionRegistry:
    def __init__(self) -> None:
        self._connections: dict[str, set[WebSocketProtocol]] = {}
        self._lock = asyncio.Lock()

    async def register(self, worker_id: str, ws: WebSocketProtocol) -> None:
        async with self._lock:
            if worker_id not in self._connections:
                self._connections[worker_id] = set()
            self._connections[worker_id].add(ws)

    async def unregister(self, worker_id: str, ws: WebSocketProtocol) -> None:
        async with self._lock:
            if worker_id in self._connections:
                self._connections[worker_id].discard(ws)
                if not self._connections[worker_id]:
                    del self._connections[worker_id]

    def get_connections(self, worker_id: str) -> set[WebSocketProtocol]:
        return self._connections.get(worker_id, set()).copy()

    def get_connection_count(self, worker_id: str) -> int:
        return len(self._connections.get(worker_id, set()))

    async def broadcast(
        self,
        worker_id: str,
        message: str,
    ) -> int:
        connections = self.get_connections(worker_id)
        if not connections:
            return 0

        results = await asyncio.gather(
            *[self._safe_send(ws, message) for ws in connections],
            return_exceptions=True,
        )
        return sum(1 for r in results if r is True)

    async def broadcast_except(
        self,
        worker_id: str,
        exclude: WebSocketProtocol,
        message: str,
    ) -> int:

        connections = self.get_connections(worker_id)
        connections.discard(exclude)
        if not connections:
            return 0

        results = await asyncio.gather(
            *[self._safe_send(ws, message) for ws in connections],
            return_exceptions=True,
        )
        return sum(1 for r in results if r is True)

    async def _safe_send(self, ws: WebSocketProtocol, message: str) -> bool:
        try:
            await ws.send_text(message)
            return True
        except Exception:
            return False



import json
import uuid

from agents import Agent, Runner, function_tool
from fastapi import FastAPI, WebSocket

from dooers import WorkerConfig, WorkerServer

app = FastAPI()
worker_server = WorkerServer(
    WorkerConfig(
        database_type="sqlite",
        database_name="worker.db",
        assistant_name="OpenAI Agent",
    )
)


@function_tool
def get_weather(city: str) -> str:
    """Get the weather for a city."""
    return f"The weather in {city} is sunny, 22C."


agent = Agent(
    name="Assistant",
    instructions="You are a helpful assistant.",
    tools=[get_weather],
)


async def openai_agents_handler(incoming, send, memory, analytics, settings):
    yield send.run_start(agent_id="openai-agents")

    # get_history_raw() returns ThreadEvent objects for manual conversion
    events = await memory.get_history_raw(limit=20, order="asc")

    input_messages = []
    for event in events:
        if event.type == "message" and event.content:
            text = " ".join(part.text for part in event.content if hasattr(part, "text"))
            if text:
                role = "user" if event.actor == "user" else "assistant"
                input_messages.append({"role": role, "content": text})

    input_messages.append({"role": "user", "content": incoming.message})

    result = await Runner.run(agent, input=input_messages)

    tool_call_ids: dict[str, str] = {}

    for item in result.new_items:
        if item.type == "tool_call_item":
            call_id = str(uuid.uuid4())
            args = json.loads(item.raw_item.arguments)
            tool_call_ids[item.raw_item.call_id] = call_id
            yield send.tool_call(
                item.raw_item.name,
                args,
                display_name=f"Calling {item.raw_item.name}...",
                id=call_id,
            )
        elif item.type == "tool_call_output_item":
            call_id = tool_call_ids.get(item.raw_item.call_id)
            yield send.tool_result(
                item.raw_item.name,
                {"output": item.output},
                id=call_id,
            )

    yield send.text(result.final_output)
    yield send.update_thread(title=incoming.message[:60])
    yield send.run_end()


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    await worker_server.handle(websocket, openai_agents_handler)


@app.on_event("shutdown")
async def shutdown():
    await worker_server.close()



class DispatchError(Exception):
    """Raised during dispatch setup phase.

    Indicates infrastructure failures (database down, invalid arguments)
    that prevent handler execution from starting. No thread is created
    and no events are broadcast.
    """


class HandlerError(Exception):
    """Raised when a handler fails during execution.

    Before this exception reaches the caller, the pipeline has already:
    1. Logged the error
    2. Marked the current run as failed
    3. Created a system error event in the thread
    4. Broadcast both to WebSocket subscribers

    The thread is always in a clean state when this exception is raised.
    """

    def __init__(self, message: str, original: Exception | None = None):
        super().__init__(message)
        self.original = original



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .collector import AnalyticsCollector


class WorkerAnalytics:
    """
    Handler API for analytics tracking.

    Provides methods for tracking custom events and recording feedback.
    Passed as the fourth parameter to handler functions.
    """

    def __init__(
        self,
        worker_id: str,
        thread_id: str,
        user_id: str | None,
        run_id: str | None,
        collector: AnalyticsCollector,
        organization_id: str | None = None,
        workspace_id: str | None = None,
    ) -> None:
        self._worker_id = worker_id
        self._thread_id = thread_id
        self._user_id = user_id
        self._run_id = run_id
        self._collector = collector
        self._organization_id = organization_id
        self._workspace_id = workspace_id

    async def track(self, event: str, data: dict[str, Any] | None = None) -> None:
        """
        Track a custom analytics event.

        Args:
            event: Event name (e.g., "llm.called", "search.performed")
            data: Optional event-specific data
        """
        await self._collector.track(
            event=event,
            worker_id=self._worker_id,
            thread_id=self._thread_id,
            user_id=self._user_id,
            run_id=self._run_id,
            data=data,
            organization_id=self._organization_id,
            workspace_id=self._workspace_id,
        )

    async def like(
        self,
        target_type: str,
        target_id: str,
        reason: str | None = None,
        classification: str | None = None,
    ) -> None:
        """
        Record a like (positive feedback).

        Useful for external sources like WhatsApp where users can't
        interact with the dashboard directly.

        Args:
            target_type: Type of target ("event", "run", or "thread")
            target_id: ID of the target
            reason: Optional reason for the feedback
            classification: Optional classification label for the feedback
        """
        await self._collector.feedback(
            feedback_type="like",
            target_type=target_type,
            target_id=target_id,
            worker_id=self._worker_id,
            thread_id=self._thread_id,
            user_id=self._user_id,
            reason=reason,
            classification=classification,
            organization_id=self._organization_id,
            workspace_id=self._workspace_id,
        )

    async def dislike(
        self,
        target_type: str,
        target_id: str,
        reason: str | None = None,
        classification: str | None = None,
    ) -> None:
        """
        Record a dislike (negative feedback).

        Useful for external sources like WhatsApp where users can't
        interact with the dashboard directly.

        Args:
            target_type: Type of target ("event", "run", or "thread")
            target_id: ID of the target
            reason: Optional reason for the feedback
            classification: Optional classification label for the feedback
        """
        await self._collector.feedback(
            feedback_type="dislike",
            target_type=target_type,
            target_id=target_id,
            worker_id=self._worker_id,
            thread_id=self._thread_id,
            user_id=self._user_id,
            reason=reason,
            classification=classification,
            organization_id=self._organization_id,
            workspace_id=self._workspace_id,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ..persistence.base import Persistence
    from .broadcaster import SettingsBroadcaster
    from .models import SettingsSchema


class WorkerSettings:
    def __init__(
        self,
        worker_id: str,
        schema: SettingsSchema,
        persistence: Persistence,
        broadcaster: SettingsBroadcaster,
    ) -> None:
        self._worker_id = worker_id
        self._schema = schema
        self._persistence = persistence
        self._broadcaster = broadcaster

    async def get(self, field_id: str) -> Any:
        field = self._schema.get_field(field_id)
        if not field:
            raise KeyError(f"Unknown field: {field_id}")

        values = await self._persistence.get_settings(self._worker_id)
        return values.get(field_id, field.value)

    async def get_all(self, *, exclude: list[str] | None = None) -> dict[str, Any]:
        defaults = self._schema.get_defaults()
        stored = await self._persistence.get_settings(self._worker_id)
        merged = {**defaults, **stored}
        if exclude:
            for key in exclude:
                merged.pop(key, None)
        return merged

    async def set(self, field_id: str, value: Any) -> None:
        self._validate_field(field_id, value)
        await self._persistence.update_setting(self._worker_id, field_id, value)
        await self._broadcaster.broadcast_patch(self._worker_id, field_id, value, schema=self._schema)

    def _validate_field(self, field_id: str, value: Any) -> None:
        field = self._schema.get_field(field_id)
        if not field:
            raise KeyError(f"Unknown field: {field_id}")
        if field.readonly:
            raise ValueError(f"Field '{field_id}' is readonly")



import os

from fastapi import FastAPI, WebSocket
from langchain_core.messages import AIMessage, HumanMessage, SystemMessage
from langchain_openai import ChatOpenAI

from dooers import WorkerConfig, WorkerServer

app = FastAPI()
worker_server = WorkerServer(
    WorkerConfig(
        database_type="sqlite",
        database_name="worker.db",
    )
)
llm = ChatOpenAI(
    model="gpt-4o-mini",
    api_key=os.getenv("OPENAI_API_KEY"),
)


async def langchain_agent(incoming, send, memory, analytics, settings):
    yield send.run_start(agent_id="langchain")

    # get_history_raw() returns ThreadEvent objects for manual conversion
    events = await memory.get_history_raw(limit=20, order="asc")

    messages = [SystemMessage(content="You are a helpful assistant.")]
    for event in events:
        if event.type == "message" and event.content:
            text = " ".join(part.text for part in event.content if hasattr(part, "text"))
            if text:
                if event.actor == "user":
                    messages.append(HumanMessage(content=text))
                else:
                    messages.append(AIMessage(content=text))

    messages.append(HumanMessage(content=incoming.message))

    result = await llm.ainvoke(messages)

    yield send.text(result.content)
    yield send.update_thread(title=incoming.message[:60])
    yield send.run_end()


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    await worker_server.handle(websocket, langchain_agent)


@app.on_event("shutdown")
async def shutdown():
    await worker_server.close()



from __future__ import annotations

import logging
import uuid
from collections.abc import AsyncGenerator, Awaitable, Callable
from dataclasses import dataclass
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any

from dooers.exceptions import HandlerError
from dooers.features.analytics.models import AnalyticsEvent
from dooers.features.analytics.worker_analytics import WorkerAnalytics
from dooers.features.settings.worker_settings import WorkerSettings
from dooers.handlers.context import WorkerContext
from dooers.handlers.incoming import WorkerIncoming
from dooers.handlers.memory import WorkerMemory
from dooers.handlers.send import WorkerEvent, WorkerSend
from dooers.persistence.base import Persistence
from dooers.protocol.models import (
    AudioPart,
    ContentPart,
    DocumentPart,
    ImagePart,
    Run,
    TextPart,
    Thread,
    ThreadEvent,
    User,
    WireC2S_ContentPart,
    WireS2C_AudioPart,
    WireS2C_ContentPart,
    WireS2C_DocumentPart,
    WireS2C_ImagePart,
    WireS2C_TextPart,
    deserialize_s2c_part,
)

if TYPE_CHECKING:
    from dooers.features.analytics.collector import AnalyticsCollector
    from dooers.features.settings.broadcaster import SettingsBroadcaster
    from dooers.features.settings.models import SettingsSchema
    from dooers.upload_store import UploadStore

logger = logging.getLogger("workers")


class UploadReferenceError(ValueError):
    """Raised when a ref_id cannot be resolved from the upload store."""


def _generate_id() -> str:
    return str(uuid.uuid4())


def _now() -> datetime:
    return datetime.now(UTC)


def _user_author_display(user: User) -> str | None:
    """Display name for persisted user message events (Cosmos/Postgres `author` field)."""
    if user.user_name and str(user.user_name).strip():
        return str(user.user_name).strip()
    if user.user_email and str(user.user_email).strip():
        return str(user.user_email).strip()
    if user.user_id and str(user.user_id).strip():
        return str(user.user_id).strip()
    return None


Handler = Callable[
    [WorkerIncoming, WorkerSend, WorkerMemory, WorkerAnalytics, WorkerSettings],
    AsyncGenerator[WorkerEvent, None],
]


@dataclass
class HandlerContext:
    handler: Handler
    worker_id: str
    message: str
    organization_id: str = ""
    workspace_id: str = ""
    user: User = None  # type: ignore[assignment]
    thread_id: str | None = None
    thread_title: str | None = None
    content: list[WireC2S_ContentPart | dict[str, Any]] | None = None
    data: dict[str, Any] | None = None
    client_event_id: str | None = None

    def __post_init__(self):
        if self.user is None:
            self.user = User(user_id="")


@dataclass
class PipelineResult:
    thread: Thread
    user_event: ThreadEvent
    is_new_thread: bool
    handler_content: list[ContentPart] | None = None


class HandlerPipeline:
    def __init__(
        self,
        persistence: Persistence,
        broadcast_callback: Callable[[str, dict[str, Any]], Awaitable[None]] | None = None,
        analytics_collector: AnalyticsCollector | None = None,
        settings_broadcaster: SettingsBroadcaster | None = None,
        settings_schema: SettingsSchema | None = None,
        assistant_name: str = "Assistant",
        upload_store: UploadStore | None = None,
    ):
        self._persistence = persistence
        self._broadcast_callback = broadcast_callback
        self._analytics_collector = analytics_collector
        self._settings_broadcaster = settings_broadcaster
        self._settings_schema = settings_schema
        self._assistant_name = assistant_name
        self._upload_store = upload_store

    async def setup(self, context: HandlerContext) -> PipelineResult:
        now = _now()
        thread_id = context.thread_id
        is_new_thread = False

        if not thread_id:
            thread_id = _generate_id()
            thread = Thread(
                id=thread_id,
                worker_id=context.worker_id,
                organization_id=context.organization_id,
                workspace_id=context.workspace_id,
                owner=context.user,
                users=[context.user],
                title=context.thread_title,
                created_at=now,
                updated_at=now,
                last_event_at=now,
            )
            await self._persistence.create_thread(thread)
            is_new_thread = True

            await self._broadcast(
                context.worker_id,
                {
                    "type": "thread.upsert",
                    "thread": thread,
                },
            )

            await self._track_event(
                context.worker_id,
                AnalyticsEvent.THREAD_CREATED.value,
                thread_id=thread_id,
                user_id=context.user.user_id,
                organization_id=context.organization_id,
                workspace_id=context.workspace_id,
            )
        else:
            thread = await self._persistence.get_thread(thread_id)
            if thread:
                if thread.worker_id != context.worker_id:
                    raise PermissionError(f"Thread {thread_id} belongs to different worker")
            else:
                # Auto-create thread for deterministic IDs (e.g., dispatch with pre-computed thread_id)
                thread = Thread(
                    id=thread_id,
                    worker_id=context.worker_id,
                    organization_id=context.organization_id,
                    workspace_id=context.workspace_id,
                    owner=context.user,
                    users=[context.user],
                    title=context.thread_title,
                    created_at=now,
                    updated_at=now,
                    last_event_at=now,
                )
                await self._persistence.create_thread(thread)
                is_new_thread = True

                await self._broadcast(
                    context.worker_id,
                    {
                        "type": "thread.upsert",
                        "thread": thread,
                    },
                )

                await self._track_event(
                    context.worker_id,
                    AnalyticsEvent.THREAD_CREATED.value,
                    thread_id=thread_id,
                    user_id=context.user.user_id,
                    organization_id=context.organization_id,
                    workspace_id=context.workspace_id,
                )

        if context.content:
            handler_parts, storage_parts = self._resolve_content_parts(context.content)
        else:
            text = context.message or ""
            handler_parts = [TextPart(text=text)]
            storage_parts = [WireS2C_TextPart(text=text)]

        user_event_id = _generate_id()
        user_event = ThreadEvent(
            id=user_event_id,
            thread_id=thread_id,
            run_id=None,
            type="message",
            actor="user",
            author=_user_author_display(context.user),
            user=context.user,
            content=storage_parts,
            data=context.data,
            created_at=now,
            client_event_id=context.client_event_id,
        )
        await self._persistence.create_event(user_event)

        await self._broadcast(
            context.worker_id,
            {
                "type": "event.append",
                "thread_id": thread_id,
                "events": [user_event],
            },
        )

        return PipelineResult(
            thread=thread,
            user_event=user_event,
            is_new_thread=is_new_thread,
            handler_content=handler_parts,
        )

    async def execute(
        self,
        context: HandlerContext,
        result: PipelineResult,
    ) -> AsyncGenerator[WorkerEvent, None]:
        thread_id = result.thread.id
        thread = result.thread

        handler_content = result.handler_content or []
        c2s_content_type = (
            getattr(handler_content[0], "type", "text") if handler_content else "text"
        )
        message = self._extract_message(handler_content)
        worker_context = WorkerContext(
            thread_id=thread_id,
            event_id=result.user_event.id,
            organization_id=context.organization_id,
            workspace_id=context.workspace_id,
            user=context.user,
            thread_title=thread.title,
            thread_created_at=thread.created_at,
        )
        incoming = WorkerIncoming(
            message=message,
            content=handler_content,
            context=worker_context,
        )
        send = WorkerSend()
        memory = WorkerMemory(thread_id=thread_id, persistence=self._persistence)

        analytics = self._create_analytics(
            worker_id=context.worker_id,
            thread_id=thread_id,
            user_id=context.user.user_id,
            organization_id=context.organization_id,
            workspace_id=context.workspace_id,
        )

        settings = self._create_settings(worker_id=context.worker_id)

        current_run_id: str | None = None
        current_agent_id: str | None = None

        try:
            async for event in context.handler(incoming, send, memory, analytics, settings):
                event_now = _now()

                if event.send_type == "run_start":
                    current_run_id = _generate_id()
                    current_agent_id = event.data.get("agent_id")
                    run = Run(
                        id=current_run_id,
                        thread_id=thread_id,
                        agent_id=current_agent_id,
                        status="running",
                        started_at=event_now,
                    )
                    await self._persistence.create_run(run)
                    await self._broadcast(
                        context.worker_id,
                        {
                            "type": "run.upsert",
                            "run": run,
                        },
                    )

                    result.user_event.run_id = current_run_id
                    await self._persistence.update_event(result.user_event)
                    await self._broadcast(
                        context.worker_id,
                        {
                            "type": "event.append",
                            "thread_id": thread_id,
                            "events": [result.user_event],
                        },
                    )
                    await self._track_event(
                        context.worker_id,
                        AnalyticsEvent.MESSAGE_C2S.value,
                        thread_id=thread_id,
                        user_id=context.user.user_id,
                        run_id=current_run_id,
                        event_id=result.user_event.id,
                        data={"type": c2s_content_type},
                        organization_id=context.organization_id,
                        workspace_id=context.workspace_id,
                    )

                elif event.send_type == "run_end":
                    if current_run_id:
                        run = Run(
                            id=current_run_id,
                            thread_id=thread_id,
                            agent_id=current_agent_id,
                            status=event.data.get("status", "succeeded"),
                            started_at=event_now,
                            ended_at=event_now,
                            error=event.data.get("error"),
                        )
                        await self._persistence.update_run(run)
                        await self._broadcast(
                            context.worker_id,
                            {
                                "type": "run.upsert",
                                "run": run,
                            },
                        )
                        current_run_id = None
                        current_agent_id = None

                elif event.send_type == "text":
                    event_id = _generate_id()
                    thread_event = ThreadEvent(
                        id=event_id,
                        thread_id=thread_id,
                        run_id=current_run_id,
                        type="message",
                        actor="assistant",
                        author=event.data.get("author") or self._assistant_name,
                        content=[WireS2C_TextPart(text=event.data["text"])],
                        created_at=event_now,
                    )
                    await self._persistence.create_event(thread_event)
                    await self._broadcast(
                        context.worker_id,
                        {
                            "type": "event.append",
                            "thread_id": thread_id,
                            "events": [thread_event],
                        },
                    )
                    await self._track_event(
                        context.worker_id,
                        AnalyticsEvent.MESSAGE_S2C.value,
                        thread_id=thread_id,
                        user_id=context.user.user_id,
                        run_id=current_run_id,
                        event_id=event_id,
                        data={"type": "text"},
                        organization_id=context.organization_id,
                        workspace_id=context.workspace_id,
                    )

                elif event.send_type == "audio":
                    event_id = _generate_id()
                    thread_event = ThreadEvent(
                        id=event_id,
                        thread_id=thread_id,
                        run_id=current_run_id,
                        type="message",
                        actor="assistant",
                        author=event.data.get("author") or self._assistant_name,
                        content=[
                            WireS2C_AudioPart(
                                url=event.data["url"],
                                mime_type=event.data.get("mime_type"),
                                duration=event.data.get("duration"),
                            )
                        ],
                        created_at=event_now,
                    )
                    await self._persistence.create_event(thread_event)
                    await self._broadcast(
                        context.worker_id,
                        {
                            "type": "event.append",
                            "thread_id": thread_id,
                            "events": [thread_event],
                        },
                    )
                    await self._track_event(
                        context.worker_id,
                        AnalyticsEvent.MESSAGE_S2C.value,
                        thread_id=thread_id,
                        user_id=context.user.user_id,
                        run_id=current_run_id,
                        event_id=event_id,
                        data={"type": "audio"},
                        organization_id=context.organization_id,
                        workspace_id=context.workspace_id,
                    )

                elif event.send_type == "image":
                    event_id = _generate_id()
                    thread_event = ThreadEvent(
                        id=event_id,
                        thread_id=thread_id,
                        run_id=current_run_id,
                        type="message",
                        actor="assistant",
                        author=event.data.get("author") or self._assistant_name,
                        content=[
                            WireS2C_ImagePart(
                                url=event.data["url"],
                                mime_type=event.data.get("mime_type"),
                                alt=event.data.get("alt"),
                            )
                        ],
                        created_at=event_now,
                    )
                    await self._persistence.create_event(thread_event)
                    await self._broadcast(
                        context.worker_id,
                        {
                            "type": "event.append",
                            "thread_id": thread_id,
                            "events": [thread_event],
                        },
                    )
                    await self._track_event(
                        context.worker_id,
                        AnalyticsEvent.MESSAGE_S2C.value,
                        thread_id=thread_id,
                        user_id=context.user.user_id,
                        run_id=current_run_id,
                        event_id=event_id,
                        data={"type": "image"},
                        organization_id=context.organization_id,
                        workspace_id=context.workspace_id,
                    )

                elif event.send_type == "document":
                    event_id = _generate_id()
                    thread_event = ThreadEvent(
                        id=event_id,
                        thread_id=thread_id,
                        run_id=current_run_id,
                        type="message",
                        actor="assistant",
                        author=event.data.get("author") or self._assistant_name,
                        content=[
                            WireS2C_DocumentPart(
                                url=event.data["url"],
                                filename=event.data["filename"],
                                mime_type=event.data["mime_type"],
                            )
                        ],
                        created_at=event_now,
                    )
                    await self._persistence.create_event(thread_event)
                    await self._broadcast(
                        context.worker_id,
                        {
                            "type": "event.append",
                            "thread_id": thread_id,
                            "events": [thread_event],
                        },
                    )
                    await self._track_event(
                        context.worker_id,
                        AnalyticsEvent.MESSAGE_S2C.value,
                        thread_id=thread_id,
                        user_id=context.user.user_id,
                        run_id=current_run_id,
                        event_id=event_id,
                        data={"type": "document"},
                        organization_id=context.organization_id,
                        workspace_id=context.workspace_id,
                    )

                elif event.send_type == "tool_call":
                    event_id = _generate_id()
                    thread_event = ThreadEvent(
                        id=event_id,
                        thread_id=thread_id,
                        run_id=current_run_id,
                        type="tool.call",
                        actor="assistant",
                        data={
                            "id": event.data["id"],
                            "name": event.data["name"],
                            "display_name": event.data.get("display_name"),
                            "args": event.data["args"],
                        },
                        created_at=event_now,
                    )
                    await self._persistence.create_event(thread_event)
                    await self._broadcast(
                        context.worker_id,
                        {
                            "type": "event.append",
                            "thread_id": thread_id,
                            "events": [thread_event],
                        },
                    )
                    await self._track_event(
                        context.worker_id,
                        AnalyticsEvent.TOOL_CALLED.value,
                        thread_id=thread_id,
                        user_id=context.user.user_id,
                        run_id=current_run_id,
                        event_id=event_id,
                        data={"name": event.data["name"]},
                        organization_id=context.organization_id,
                        workspace_id=context.workspace_id,
                    )

                elif event.send_type == "tool_result":
                    thread_event = ThreadEvent(
                        id=_generate_id(),
                        thread_id=thread_id,
                        run_id=current_run_id,
                        type="tool.result",
                        actor="tool",
                        data={
                            "id": event.data.get("id"),
                            "name": event.data["name"],
                            "display_name": event.data.get("display_name"),
                            "args": event.data.get("args"),
                            "result": event.data["result"],
                        },
                        created_at=event_now,
                    )
                    await self._persistence.create_event(thread_event)
                    await self._broadcast(
                        context.worker_id,
                        {
                            "type": "event.append",
                            "thread_id": thread_id,
                            "events": [thread_event],
                        },
                    )

                elif event.send_type == "tool_transaction":
                    event_id = _generate_id()
                    thread_event = ThreadEvent(
                        id=event_id,
                        thread_id=thread_id,
                        run_id=current_run_id,
                        type="tool.transaction",
                        actor="assistant",
                        data={
                            "id": event.data["id"],
                            "name": event.data["name"],
                            "display_name": event.data.get("display_name"),
                            "args": event.data["args"],
                            "result": event.data["result"],
                        },
                        created_at=event_now,
                    )
                    await self._persistence.create_event(thread_event)
                    await self._broadcast(
                        context.worker_id,
                        {
                            "type": "event.append",
                            "thread_id": thread_id,
                            "events": [thread_event],
                        },
                    )
                    await self._track_event(
                        context.worker_id,
                        AnalyticsEvent.TOOL_CALLED.value,
                        thread_id=thread_id,
                        user_id=context.user.user_id,
                        run_id=current_run_id,
                        event_id=event_id,
                        data={"name": event.data["name"]},
                        organization_id=context.organization_id,
                        workspace_id=context.workspace_id,
                    )

                elif event.send_type == "event_update":
                    target_event_id = event.data["event_id"]
                    existing_event = await self._persistence.get_event(target_event_id)
                    if not existing_event:
                        logger.warning("[workers] event_update: event %s not found", target_event_id)
                    else:
                        existing_event.content = [deserialize_s2c_part(p) for p in event.data["content"]]
                        await self._persistence.update_event(existing_event)
                        await self._broadcast(
                            context.worker_id,
                            {
                                "type": "event.append",
                                "thread_id": existing_event.thread_id,
                                "events": [existing_event],
                            },
                        )

                elif event.send_type == "thread_update":
                    thread = await self._persistence.get_thread(thread_id)
                    if thread:
                        if event.data.get("title") is not None:
                            thread.title = event.data["title"]
                        thread.updated_at = event_now
                        await self._persistence.update_thread(thread)
                        await self._broadcast(
                            context.worker_id,
                            {
                                "type": "thread.upsert",
                                "thread": thread,
                            },
                        )

                if event.send_type != "thread_update":
                    await self._update_thread_last_event(thread_id, event_now)

                yield event

            if result.user_event.run_id is None:
                await self._track_event(
                    context.worker_id,
                    AnalyticsEvent.MESSAGE_C2S.value,
                    thread_id=thread_id,
                    user_id=context.user.user_id,
                    run_id=None,
                    event_id=result.user_event.id,
                    data={"type": c2s_content_type},
                    organization_id=context.organization_id,
                    workspace_id=context.workspace_id,
                )

        except Exception as e:
            if result.user_event.run_id is None:
                await self._track_event(
                    context.worker_id,
                    AnalyticsEvent.MESSAGE_C2S.value,
                    thread_id=thread_id,
                    user_id=context.user.user_id,
                    run_id=None,
                    event_id=result.user_event.id,
                    data={"type": c2s_content_type},
                    organization_id=context.organization_id,
                    workspace_id=context.workspace_id,
                )

            logger.error("[workers] handler error: %s", e, exc_info=True)

            await self._track_event(
                context.worker_id,
                AnalyticsEvent.ERROR_OCCURRED.value,
                thread_id=thread_id,
                user_id=context.user.user_id,
                run_id=current_run_id,
                data={"error": str(e), "type": type(e).__name__},
                organization_id=context.organization_id,
                workspace_id=context.workspace_id,
            )

            error_now = _now()

            if current_run_id:
                run = Run(
                    id=current_run_id,
                    thread_id=thread_id,
                    agent_id=current_agent_id,
                    status="failed",
                    started_at=error_now,
                    ended_at=error_now,
                    error=str(e),
                )
                await self._persistence.update_run(run)
                await self._broadcast(
                    context.worker_id,
                    {
                        "type": "run.upsert",
                        "run": run,
                    },
                )

            error_event = ThreadEvent(
                id=_generate_id(),
                thread_id=thread_id,
                run_id=current_run_id,
                type="message",
                actor="system",
                content=[WireS2C_TextPart(text=str(e))],
                created_at=error_now,
            )
            await self._persistence.create_event(error_event)
            await self._broadcast(
                context.worker_id,
                {
                    "type": "event.append",
                    "thread_id": thread_id,
                    "events": [error_event],
                },
            )
            await self._update_thread_last_event(thread_id, error_now)

            raise HandlerError(str(e), original=e) from e

    async def _broadcast(self, worker_id: str, payload: dict[str, Any]) -> None:
        if self._broadcast_callback:
            await self._broadcast_callback(worker_id, payload)

    async def _track_event(
        self,
        worker_id: str,
        event: str,
        *,
        thread_id: str | None = None,
        user_id: str | None = None,
        run_id: str | None = None,
        event_id: str | None = None,
        data: dict[str, Any] | None = None,
        organization_id: str | None = None,
        workspace_id: str | None = None,
    ) -> None:
        if self._analytics_collector:
            await self._analytics_collector.track(
                event=event,
                worker_id=worker_id,
                thread_id=thread_id,
                user_id=user_id,
                run_id=run_id,
                event_id=event_id,
                data=data,
                organization_id=organization_id,
                workspace_id=workspace_id,
            )

    async def _update_thread_last_event(self, thread_id: str, timestamp: datetime) -> None:
        thread = await self._persistence.get_thread(thread_id)
        if thread:
            thread.last_event_at = timestamp
            thread.updated_at = timestamp
            await self._persistence.update_thread(thread)

    def _create_analytics(
        self,
        worker_id: str,
        thread_id: str,
        user_id: str | None,
        organization_id: str | None,
        workspace_id: str | None,
    ) -> WorkerAnalytics:
        if self._analytics_collector:
            return WorkerAnalytics(
                worker_id=worker_id,
                thread_id=thread_id,
                user_id=user_id,
                run_id=None,
                collector=self._analytics_collector,
                organization_id=organization_id,
                workspace_id=workspace_id,
            )

        class NoopCollector:
            async def track(self, **kwargs) -> None:
                pass

            async def feedback(self, **kwargs) -> None:
                pass

        return WorkerAnalytics(
            worker_id=worker_id,
            thread_id=thread_id,
            user_id=user_id,
            run_id=None,
            collector=NoopCollector(),  # type: ignore
            organization_id=organization_id,
            workspace_id=workspace_id,
        )

    def _create_settings(self, worker_id: str) -> WorkerSettings:
        if self._settings_schema and self._settings_broadcaster:
            return WorkerSettings(
                worker_id=worker_id,
                schema=self._settings_schema,
                persistence=self._persistence,
                broadcaster=self._settings_broadcaster,
            )

        from dooers.features.settings.models import SettingsSchema

        class NoopBroadcaster:
            async def broadcast_snapshot(self, **kwargs) -> None:
                pass

            async def broadcast_patch(self, **kwargs) -> None:
                pass

        class NoopPersistence:
            async def get_settings(self, worker_id: str) -> dict:
                return {}

            async def update_setting(self, worker_id: str, field_id: str, value) -> None:
                pass

            async def set_settings(self, worker_id: str, values: dict) -> None:
                pass

        return WorkerSettings(
            worker_id=worker_id,
            schema=SettingsSchema(fields=[]),
            persistence=NoopPersistence(),  # type: ignore
            broadcaster=NoopBroadcaster(),  # type: ignore
        )

    def _resolve_content_parts(
        self, parts: list[WireC2S_ContentPart | dict[str, Any]]
    ) -> tuple[list[ContentPart], list[WireS2C_ContentPart]]:
        """Resolve C2S wire content parts into handler format (with bytes) and
        storage format (metadata only, no bytes).

        Returns (handler_parts, storage_parts).
        """
        handler_parts: list[ContentPart] = []
        storage_parts: list[WireS2C_ContentPart] = []

        for part in parts:
            if hasattr(part, "model_dump"):
                data = part.model_dump()
            elif isinstance(part, dict):
                data = part
            else:
                raise ValueError(f"Unsupported content part: {type(part)}")

            part_type = data.get("type")

            if part_type == "text":
                handler_parts.append(TextPart(text=data["text"]))
                storage_parts.append(WireS2C_TextPart(text=data["text"]))

            elif part_type in ("audio", "image", "document") and "ref_id" in data:
                # WebSocket path — resolve ref_id from upload store
                ref_id = data["ref_id"]
                entry = self._upload_store.consume(ref_id) if self._upload_store else None
                if entry is None:
                    raise UploadReferenceError(f"Upload reference '{ref_id}' not found or expired")

                if part_type == "audio":
                    handler_parts.append(
                        AudioPart(
                            data=entry.data,
                            mime_type=entry.mime_type,
                            duration=data.get("duration"),
                            filename=entry.filename,
                        )
                    )
                    storage_parts.append(
                        WireS2C_AudioPart(
                            mime_type=entry.mime_type,
                            duration=data.get("duration"),
                            filename=entry.filename,
                        )
                    )
                elif part_type == "image":
                    handler_parts.append(
                        ImagePart(
                            data=entry.data,
                            mime_type=entry.mime_type,
                            filename=entry.filename,
                        )
                    )
                    storage_parts.append(
                        WireS2C_ImagePart(
                            mime_type=entry.mime_type,
                            filename=entry.filename,
                        )
                    )
                elif part_type == "document":
                    handler_parts.append(
                        DocumentPart(
                            data=entry.data,
                            mime_type=entry.mime_type,
                            filename=entry.filename,
                            size_bytes=entry.size_bytes,
                        )
                    )
                    storage_parts.append(
                        WireS2C_DocumentPart(
                            mime_type=entry.mime_type,
                            filename=entry.filename,
                            size_bytes=entry.size_bytes,
                        )
                    )

            elif part_type == "audio" and "data" in data:
                # Dispatch path — bytes passed directly
                handler_parts.append(
                    AudioPart(
                        data=data["data"],
                        mime_type=data.get("mime_type", ""),
                        duration=data.get("duration"),
                        filename=data.get("filename"),
                    )
                )
                storage_parts.append(
                    WireS2C_AudioPart(
                        mime_type=data.get("mime_type"),
                        duration=data.get("duration"),
                        filename=data.get("filename"),
                    )
                )

            elif part_type == "image" and "data" in data:
                handler_parts.append(
                    ImagePart(
                        data=data["data"],
                        mime_type=data.get("mime_type", ""),
                        filename=data.get("filename"),
                    )
                )
                storage_parts.append(
                    WireS2C_ImagePart(
                        mime_type=data.get("mime_type"),
                        filename=data.get("filename"),
                    )
                )

            elif part_type == "document" and "data" in data:
                handler_parts.append(
                    DocumentPart(
                        data=data["data"],
                        mime_type=data.get("mime_type", ""),
                        filename=data.get("filename", ""),
                        size_bytes=data.get("size_bytes", 0),
                    )
                )
                storage_parts.append(
                    WireS2C_DocumentPart(
                        mime_type=data.get("mime_type"),
                        filename=data.get("filename"),
                        size_bytes=data.get("size_bytes"),
                    )
                )

            else:
                raise ValueError(f"Unsupported content part type: {part_type!r}")

        return handler_parts, storage_parts

    def _extract_message(self, content: list[ContentPart]) -> str:
        texts = []
        for part in content:
            if isinstance(part, TextPart):
                texts.append(part.text)
        return " ".join(texts)



from datetime import datetime
from typing import Annotated, Literal

from pydantic import BaseModel, Field

Actor = Literal["user", "assistant", "system", "tool"]
EventType = Literal["message", "run.started", "run.finished", "tool.call", "tool.result", "tool.transaction"]
RunStatus = Literal["running", "succeeded", "failed", "canceled"]


# --- Handler format (what the agent developer receives in incoming.content) ---


class TextPart(BaseModel):
    type: Literal["text"] = "text"
    text: str


class AudioPart(BaseModel):
    type: Literal["audio"] = "audio"
    data: bytes
    mime_type: str
    duration: float | None = None
    filename: str | None = None


class ImagePart(BaseModel):
    type: Literal["image"] = "image"
    data: bytes
    mime_type: str
    width: int | None = None
    height: int | None = None
    filename: str | None = None


class DocumentPart(BaseModel):
    type: Literal["document"] = "document"
    data: bytes
    mime_type: str
    filename: str
    size_bytes: int


ContentPart = Annotated[
    TextPart | AudioPart | ImagePart | DocumentPart,
    Field(discriminator="type"),
]


# --- Wire C2S format (ref_id from client uploads) ---


class WireC2S_TextPart(BaseModel):
    type: Literal["text"] = "text"
    text: str


class WireC2S_AudioPart(BaseModel):
    type: Literal["audio"] = "audio"
    ref_id: str
    duration: float | None = None


class WireC2S_ImagePart(BaseModel):
    type: Literal["image"] = "image"
    ref_id: str


class WireC2S_DocumentPart(BaseModel):
    type: Literal["document"] = "document"
    ref_id: str


WireC2S_ContentPart = Annotated[
    WireC2S_TextPart | WireC2S_AudioPart | WireC2S_ImagePart | WireC2S_DocumentPart,
    Field(discriminator="type"),
]


# --- Wire S2C format (URLs/metadata for client display) ---


class WireS2C_TextPart(BaseModel):
    type: Literal["text"] = "text"
    text: str


class WireS2C_AudioPart(BaseModel):
    type: Literal["audio"] = "audio"
    url: str | None = None
    mime_type: str | None = None
    duration: float | None = None
    filename: str | None = None


class WireS2C_ImagePart(BaseModel):
    type: Literal["image"] = "image"
    url: str | None = None
    mime_type: str | None = None
    width: int | None = None
    height: int | None = None
    alt: str | None = None
    filename: str | None = None


class WireS2C_DocumentPart(BaseModel):
    type: Literal["document"] = "document"
    url: str | None = None
    filename: str | None = None
    mime_type: str | None = None
    size_bytes: int | None = None


WireS2C_ContentPart = Annotated[
    WireS2C_TextPart | WireS2C_AudioPart | WireS2C_ImagePart | WireS2C_DocumentPart,
    Field(discriminator="type"),
]

_S2C_TYPE_MAP: dict[str, type] = {
    "text": WireS2C_TextPart,
    "audio": WireS2C_AudioPart,
    "image": WireS2C_ImagePart,
    "document": WireS2C_DocumentPart,
}


def deserialize_s2c_part(data: dict) -> WireS2C_ContentPart:
    """Deserialize a dict into a typed WireS2C content part."""
    cls = _S2C_TYPE_MAP.get(data.get("type", ""))
    if cls is None:
        raise ValueError(f"Unknown content part type: {data.get('type')!r}")
    return cls(**data)


class User(BaseModel):
    user_id: str
    user_name: str | None = None
    user_email: str | None = None
    identity_ids: list[str] = []
    system_role: str = "user"
    organization_role: str = "member"
    workspace_role: str = "member"


class ThreadEvent(BaseModel):
    id: str
    thread_id: str
    run_id: str | None = None
    type: EventType
    actor: Actor
    author: str | None = None
    user: User = User(user_id="")
    content: list[WireS2C_ContentPart] | None = None
    data: dict | None = None
    created_at: datetime
    streaming: bool | None = None
    finalized: bool | None = None
    client_event_id: str | None = None


class Thread(BaseModel):
    id: str
    worker_id: str
    organization_id: str = ""
    workspace_id: str = ""
    owner: User = User(user_id="")
    users: list[User] = []
    title: str | None = None
    created_at: datetime
    updated_at: datetime
    last_event_at: datetime


class Run(BaseModel):
    id: str
    thread_id: str
    agent_id: str | None = None
    status: RunStatus
    started_at: datetime
    ended_at: datetime | None = None
    error: str | None = None



from __future__ import annotations

import logging
import uuid
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any, Protocol

from dooers.exceptions import HandlerError
from dooers.handlers.pipeline import Handler, HandlerContext, HandlerPipeline, UploadReferenceError
from dooers.persistence.base import Persistence
from dooers.protocol.frames import (
    AckPayload,
    C2S_AnalyticsSubscribe,
    C2S_AnalyticsUnsubscribe,
    C2S_Connect,
    C2S_EventCreate,
    C2S_EventList,
    C2S_Feedback,
    C2S_SettingsPatch,
    C2S_SettingsSubscribe,
    C2S_SettingsUnsubscribe,
    C2S_ThreadDelete,
    C2S_ThreadList,
    C2S_ThreadSubscribe,
    C2S_ThreadUnsubscribe,
    ClientToServer,
    EventAppendPayload,
    EventListResultPayload,
    FeedbackAckPayload,
    RunUpsertPayload,
    S2C_Ack,
    S2C_EventAppend,
    S2C_EventListResult,
    S2C_FeedbackAck,
    S2C_RunUpsert,
    S2C_ThreadDeleted,
    S2C_ThreadListResult,
    S2C_ThreadSnapshot,
    S2C_ThreadUpsert,
    ServerToClient,
    ThreadDeletedPayload,
    ThreadListResultPayload,
    ThreadSnapshotPayload,
    ThreadUpsertPayload,
)
from dooers.protocol.models import User
from dooers.protocol.parser import serialize_frame
from dooers.registry import ConnectionRegistry

if TYPE_CHECKING:
    from dooers.features.analytics.collector import AnalyticsCollector
    from dooers.features.settings.broadcaster import SettingsBroadcaster
    from dooers.features.settings.models import SettingsSchema
    from dooers.upload_store import UploadStore

logger = logging.getLogger("workers")


class WebSocketProtocol(Protocol):
    async def receive_text(self) -> str: ...
    async def send_text(self, data: str) -> None: ...
    async def close(self, code: int = 1000) -> None: ...


def resolve_scope(user: User) -> str:
    if user.system_role == "admin":
        return "admin"
    if user.organization_role in ("owner", "manager"):
        return "organization"
    if user.workspace_role == "manager":
        return "workspace"
    return "member"


def _generate_id() -> str:
    return str(uuid.uuid4())


def _now() -> datetime:
    return datetime.now(UTC)


class Router:
    def __init__(
        self,
        persistence: Persistence,
        handler: Handler,
        registry: ConnectionRegistry,
        subscriptions: dict[str, set[str]],
        analytics_collector: AnalyticsCollector | None = None,
        settings_broadcaster: SettingsBroadcaster | None = None,
        settings_schema: SettingsSchema | None = None,
        assistant_name: str = "Assistant",
        analytics_subscriptions: dict[str, set[str]] | None = None,
        settings_subscriptions: dict[str, set[str]] | None = None,
        upload_store: UploadStore | None = None,
    ):
        self._persistence = persistence
        self._handler = handler
        self._registry = registry
        self._subscriptions = subscriptions

        self._analytics_collector = analytics_collector
        self._settings_broadcaster = settings_broadcaster
        self._settings_schema = settings_schema
        self._assistant_name = assistant_name
        self._analytics_subscriptions = analytics_subscriptions if analytics_subscriptions is not None else {}
        self._settings_subscriptions = settings_subscriptions if settings_subscriptions is not None else {}

        self._ws: WebSocketProtocol | None = None
        self._ws_id: str = _generate_id()
        self._worker_id: str | None = None
        self._user: User | None = None
        self._organization_id: str = ""
        self._workspace_id: str = ""
        self._subscribed_threads: set[str] = set()

        self._pipeline = HandlerPipeline(
            persistence=persistence,
            broadcast_callback=self._broadcast_to_worker_dict,
            analytics_collector=analytics_collector,
            settings_broadcaster=settings_broadcaster,
            settings_schema=settings_schema,
            assistant_name=assistant_name,
            upload_store=upload_store,
        )

    async def _send(self, ws: WebSocketProtocol, frame: ServerToClient) -> None:
        await ws.send_text(serialize_frame(frame))

    async def _send_ack(
        self,
        ws: WebSocketProtocol,
        ack_id: str,
        ok: bool = True,
        error: dict[str, str] | None = None,
    ) -> None:
        frame = S2C_Ack(
            id=_generate_id(),
            payload=AckPayload(ack_id=ack_id, ok=ok, error=error),
        )
        await self._send(ws, frame)

    async def _broadcast_to_worker(self, frame: ServerToClient) -> None:
        """Broadcast a frame to all connections for the current worker."""
        if not self._worker_id:
            return
        message = serialize_frame(frame)
        await self._registry.broadcast(self._worker_id, message)

    async def _broadcast_to_worker_except_self(self, ws: WebSocketProtocol, frame: ServerToClient) -> None:
        """Broadcast a frame to all connections for the current worker except this one."""
        if not self._worker_id:
            return
        message = serialize_frame(frame)
        await self._registry.broadcast_except(self._worker_id, ws, message)

    async def _broadcast_to_worker_dict(self, worker_id: str, payload: dict[str, Any]) -> None:
        """Convert a dict payload from the pipeline into S2C frames and broadcast."""
        payload_type = payload.get("type")

        if payload_type == "thread.upsert":
            frame = S2C_ThreadUpsert(
                id=_generate_id(),
                payload=ThreadUpsertPayload(thread=payload["thread"]),
            )
        elif payload_type == "event.append":
            frame = S2C_EventAppend(
                id=_generate_id(),
                payload=EventAppendPayload(
                    thread_id=payload["thread_id"],
                    events=payload["events"],
                ),
            )
        elif payload_type == "run.upsert":
            frame = S2C_RunUpsert(
                id=_generate_id(),
                payload=RunUpsertPayload(run=payload["run"]),
            )
        else:
            return

        message = serialize_frame(frame)
        await self._registry.broadcast(worker_id, message)

    async def route(self, ws: WebSocketProtocol, frame: ClientToServer) -> None:
        self._ws = ws
        match frame:
            case C2S_Connect():
                await self._handle_connect(ws, frame)
            case C2S_ThreadList():
                await self._handle_thread_list(ws, frame)
            case C2S_ThreadSubscribe():
                await self._handle_thread_subscribe(ws, frame)
            case C2S_ThreadUnsubscribe():
                await self._handle_thread_unsubscribe(ws, frame)
            case C2S_ThreadDelete():
                await self._handle_thread_delete(ws, frame)
            case C2S_EventCreate():
                await self._handle_event_create(ws, frame)
            case C2S_EventList():
                await self._handle_event_list(ws, frame)

            case C2S_AnalyticsSubscribe():
                await self._handle_analytics_subscribe(ws, frame)
            case C2S_AnalyticsUnsubscribe():
                await self._handle_analytics_unsubscribe(ws, frame)
            case C2S_Feedback():
                await self._handle_feedback(ws, frame)

            case C2S_SettingsSubscribe():
                await self._handle_settings_subscribe(ws, frame)
            case C2S_SettingsUnsubscribe():
                await self._handle_settings_unsubscribe(ws, frame)
            case C2S_SettingsPatch():
                await self._handle_settings_patch(ws, frame)

    async def cleanup(self) -> None:
        logger.info(
            "[workers] router cleanup: ws=%s, worker=%s, had_analytics_sub=%s",
            self._ws_id,
            self._worker_id,
            self._ws_id in self._analytics_subscriptions.get(self._worker_id or "", set()),
        )
        if self._worker_id:
            await self._registry.unregister(self._worker_id, self._ws)

            if self._worker_id in self._analytics_subscriptions:
                self._analytics_subscriptions[self._worker_id].discard(self._ws_id)
                if not self._analytics_subscriptions[self._worker_id]:
                    logger.info("[workers] analytics subscriptions emptied for worker %s — deleting key", self._worker_id)
                    del self._analytics_subscriptions[self._worker_id]

            if self._worker_id in self._settings_subscriptions:
                self._settings_subscriptions[self._worker_id].discard(self._ws_id)
                if not self._settings_subscriptions[self._worker_id]:
                    del self._settings_subscriptions[self._worker_id]

        if self._ws_id in self._subscriptions:
            del self._subscriptions[self._ws_id]

    async def _handle_connect(self, ws: WebSocketProtocol, frame: C2S_Connect) -> None:
        self._worker_id = frame.payload.worker_id
        self._organization_id = frame.payload.organization_id
        self._workspace_id = frame.payload.workspace_id
        # NOTE: Roles are currently trusted from the client. A future iteration should
        # validate them server-side using the auth_token (frame.payload.auth_token).
        self._user = frame.payload.user
        self._ws = ws

        await self._registry.register(self._worker_id, ws)

        self._subscriptions[self._ws_id] = set()

        await self._send_ack(ws, frame.id)

    async def _handle_thread_list(self, ws: WebSocketProtocol, frame: C2S_ThreadList) -> None:
        if not self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_CONNECTED", "message": "Must connect first"},
            )
            return

        user = self._user or User(user_id="")
        scope = resolve_scope(user)
        # Build complete identity set for thread filtering
        all_identity_ids = [uid for uid in [user.user_id, *(user.identity_ids or [])] if uid]
        limit = frame.payload.limit or 30
        threads = await self._persistence.list_threads(
            worker_id=self._worker_id,
            organization_id=self._organization_id,
            workspace_id=self._workspace_id,
            user_id=user.user_id,
            cursor=frame.payload.cursor,
            limit=limit + 1,
            scope=scope,
            user_email=user.user_email,
            identity_ids=all_identity_ids or None,
        )
        has_more = len(threads) > limit
        if has_more:
            threads = threads[:limit]

        if frame.payload.cursor:
            total_count = 0
        else:
            total_count = await self._persistence.count_threads(
                worker_id=self._worker_id,
                organization_id=self._organization_id,
                workspace_id=self._workspace_id,
                user_id=user.user_id,
                scope=scope,
                user_email=user.user_email,
                identity_ids=all_identity_ids or None,
            )

        last = threads[-1] if has_more else None
        cursor = f"{last.last_event_at.isoformat()}|{last.id}" if last else None
        result = S2C_ThreadListResult(
            id=_generate_id(),
            payload=ThreadListResultPayload(
                threads=threads,
                cursor=cursor,
                total_count=total_count,
            ),
        )
        await self._send(ws, result)

    async def _handle_thread_subscribe(
        self,
        ws: WebSocketProtocol,
        frame: C2S_ThreadSubscribe,
    ) -> None:
        if not self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_CONNECTED", "message": "Must connect first"},
            )
            return

        thread_id = frame.payload.thread_id
        thread = await self._persistence.get_thread(thread_id)

        if not thread:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_FOUND", "message": "Thread not found"},
            )
            return

        if thread.worker_id != self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "FORBIDDEN", "message": "Thread belongs to different worker"},
            )
            return

        events = await self._persistence.get_events(
            thread_id,
            after_event_id=frame.payload.after_event_id,
            limit=100,
        )

        self._subscribed_threads.add(thread_id)
        self._subscriptions[self._ws_id].add(thread_id)

        snapshot = S2C_ThreadSnapshot(
            id=_generate_id(),
            payload=ThreadSnapshotPayload(thread=thread, events=events),
        )
        await self._send(ws, snapshot)

    async def _handle_thread_unsubscribe(
        self,
        ws: WebSocketProtocol,
        frame: C2S_ThreadUnsubscribe,
    ) -> None:
        thread_id = frame.payload.thread_id
        self._subscribed_threads.discard(thread_id)
        if self._ws_id in self._subscriptions:
            self._subscriptions[self._ws_id].discard(thread_id)
        await self._send_ack(ws, frame.id)

    async def _handle_thread_delete(self, ws: WebSocketProtocol, frame: C2S_ThreadDelete) -> None:
        if not self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_CONNECTED", "message": "Must connect first"},
            )
            return

        thread_id = frame.payload.thread_id
        thread = await self._persistence.get_thread(thread_id)

        if not thread:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_FOUND", "message": "Thread not found"},
            )
            return

        if thread.worker_id != self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "FORBIDDEN", "message": "Thread belongs to different worker"},
            )
            return

        await self._persistence.delete_thread(thread_id)

        self._subscribed_threads.discard(thread_id)
        if self._ws_id in self._subscriptions:
            self._subscriptions[self._ws_id].discard(thread_id)

        deleted_frame = S2C_ThreadDeleted(
            id=_generate_id(),
            payload=ThreadDeletedPayload(thread_id=thread_id),
        )
        await self._broadcast_to_worker(deleted_frame)

        await self._send_ack(ws, frame.id)

    async def _handle_event_create(self, ws: WebSocketProtocol, frame: C2S_EventCreate) -> None:
        if not self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_CONNECTED", "message": "Must connect first"},
            )
            return

        content_parts = list(frame.payload.event.content)

        user = self._user or User(user_id="")

        context = HandlerContext(
            handler=self._handler,
            worker_id=self._worker_id,
            message="",
            organization_id=self._organization_id,
            workspace_id=self._workspace_id,
            user=user,
            thread_id=frame.payload.thread_id,
            content=content_parts,
            data=frame.payload.event.data,
            client_event_id=frame.payload.client_event_id,
        )

        try:
            result = await self._pipeline.setup(context)
        except UploadReferenceError as e:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "UPLOAD_NOT_FOUND", "message": str(e)},
            )
            return
        except ValueError:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_FOUND", "message": "Thread not found"},
            )
            return
        except PermissionError:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "FORBIDDEN", "message": "Thread belongs to different worker"},
            )
            return

        if result.is_new_thread:
            self._subscribed_threads.add(result.thread.id)
            self._subscriptions[self._ws_id].add(result.thread.id)

        # Upsert participant into thread's users array
        if user.user_id or user.user_email:
            await self._persistence.upsert_thread_participant(result.thread.id, user)

        await self._send_ack(ws, frame.id)

        try:
            async for _event in self._pipeline.execute(context, result):
                pass
        except HandlerError:
            pass  # Pipeline already handled cleanup (error event, run failure, broadcast)

    async def _handle_event_list(self, ws: WebSocketProtocol, frame: C2S_EventList) -> None:
        if not self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_CONNECTED", "message": "Must connect first"},
            )
            return

        payload = frame.payload
        limit = payload.limit
        events = await self._persistence.get_events(
            payload.thread_id,
            before_event_id=payload.before_event_id,
            limit=limit + 1,
            order="desc",
        )

        has_more = len(events) > limit
        if has_more:
            events = events[:limit]

        # Reverse to chronological order
        events.reverse()

        cursor = events[0].id if events and has_more else None
        result = S2C_EventListResult(
            id=_generate_id(),
            payload=EventListResultPayload(
                thread_id=payload.thread_id,
                events=events,
                cursor=cursor,
                has_more=has_more,
            ),
        )
        await self._send(ws, result)

    async def _handle_analytics_subscribe(
        self,
        ws: WebSocketProtocol,
        frame: C2S_AnalyticsSubscribe,
    ) -> None:
        if not self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_CONNECTED", "message": "Must connect first"},
            )
            return

        worker_id = frame.payload.worker_id
        if worker_id != self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "FORBIDDEN", "message": "Cannot subscribe to other worker's analytics"},
            )
            return

        if worker_id not in self._analytics_subscriptions:
            self._analytics_subscriptions[worker_id] = set()
        self._analytics_subscriptions[worker_id].add(self._ws_id)

        logger.info(
            "[workers] analytics subscribed: ws=%s, worker=%s, total_subscribers=%d",
            self._ws_id,
            worker_id,
            len(self._analytics_subscriptions[worker_id]),
        )

        await self._send_ack(ws, frame.id)

    async def _handle_analytics_unsubscribe(
        self,
        ws: WebSocketProtocol,
        frame: C2S_AnalyticsUnsubscribe,
    ) -> None:
        worker_id = frame.payload.worker_id
        logger.info(
            "[workers] analytics unsubscribed: ws=%s, worker=%s",
            self._ws_id,
            worker_id,
        )
        if worker_id in self._analytics_subscriptions:
            self._analytics_subscriptions[worker_id].discard(self._ws_id)
            if not self._analytics_subscriptions[worker_id]:
                del self._analytics_subscriptions[worker_id]

        await self._send_ack(ws, frame.id)

    async def _handle_feedback(
        self,
        ws: WebSocketProtocol,
        frame: C2S_Feedback,
    ) -> None:
        if not self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_CONNECTED", "message": "Must connect first"},
            )
            return

        if self._analytics_collector:
            user = self._user or User(user_id="")
            # Derive thread_id only when the feedback directly targets a thread.
            thread_id = frame.payload.target_id if frame.payload.target_type == "thread" else None
            await self._analytics_collector.feedback(
                feedback_type=frame.payload.feedback,
                target_type=frame.payload.target_type,
                target_id=frame.payload.target_id,
                worker_id=self._worker_id,
                user_id=user.user_id,
                reason=frame.payload.reason,
                classification=frame.payload.classification,
                thread_id=thread_id,
                organization_id=self._organization_id,
                workspace_id=self._workspace_id,
            )

        ack = S2C_FeedbackAck(
            id=_generate_id(),
            payload=FeedbackAckPayload(
                target_type=frame.payload.target_type,
                target_id=frame.payload.target_id,
                feedback=frame.payload.feedback,
                ok=True,
            ),
        )
        await self._send(ws, ack)

    async def _handle_settings_subscribe(
        self,
        ws: WebSocketProtocol,
        frame: C2S_SettingsSubscribe,
    ) -> None:
        if not self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_CONNECTED", "message": "Must connect first"},
            )
            return

        worker_id = frame.payload.worker_id
        if worker_id != self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "FORBIDDEN", "message": "Cannot subscribe to other worker's settings"},
            )
            return

        if not self._settings_schema:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_CONFIGURED", "message": "No settings schema configured"},
            )
            return

        if worker_id not in self._settings_subscriptions:
            self._settings_subscriptions[worker_id] = set()
        self._settings_subscriptions[worker_id].add(self._ws_id)

        if self._settings_broadcaster:
            values = await self._persistence.get_settings(worker_id)
            await self._settings_broadcaster.broadcast_snapshot_to_ws(
                worker_id=worker_id,
                ws=ws,
                schema=self._settings_schema,
                values=values,
            )

        await self._send_ack(ws, frame.id)

    async def _handle_settings_unsubscribe(
        self,
        ws: WebSocketProtocol,
        frame: C2S_SettingsUnsubscribe,
    ) -> None:
        worker_id = frame.payload.worker_id
        if worker_id in self._settings_subscriptions:
            self._settings_subscriptions[worker_id].discard(self._ws_id)
            if not self._settings_subscriptions[worker_id]:
                del self._settings_subscriptions[worker_id]

        await self._send_ack(ws, frame.id)

    async def _handle_settings_patch(
        self,
        ws: WebSocketProtocol,
        frame: C2S_SettingsPatch,
    ) -> None:
        if not self._worker_id:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_CONNECTED", "message": "Must connect first"},
            )
            return

        if not self._settings_schema:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_CONFIGURED", "message": "No settings schema configured"},
            )
            return

        field_id = frame.payload.field_id
        value = frame.payload.value

        field = self._settings_schema.get_field(field_id)
        if not field:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "NOT_FOUND", "message": f"Unknown field: {field_id}"},
            )
            return

        if field.readonly:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "READONLY", "message": f"Field '{field_id}' is readonly"},
            )
            return

        if field.is_internal:
            await self._send_ack(
                ws,
                frame.id,
                ok=False,
                error={"code": "INTERNAL", "message": f"Field '{field_id}' is internal"},
            )
            return

        await self._persistence.update_setting(self._worker_id, field_id, value)

        if self._settings_broadcaster:
            await self._settings_broadcaster.broadcast_patch(
                worker_id=self._worker_id,
                field_id=field_id,
                value=value,
            )

        await self._send_ack(ws, frame.id)



from dooers.broadcast import BroadcastManager
from dooers.config import WorkerConfig
from dooers.dispatch import DispatchStream
from dooers.exceptions import DispatchError, HandlerError
from dooers.features.analytics import (
    AnalyticsBatch,
    AnalyticsCollector,
    AnalyticsEvent,
    AnalyticsEventPayload,
    WorkerAnalytics,
)
from dooers.features.settings import (
    SettingsBroadcaster,
    SettingsField,
    SettingsFieldGroup,
    SettingsFieldType,
    SettingsSchema,
    SettingsSelectOption,
    WorkerSettings,
)
from dooers.handlers.context import WorkerContext
from dooers.handlers.incoming import WorkerIncoming
from dooers.handlers.memory import WorkerMemory
from dooers.handlers.pipeline import Handler, UploadReferenceError
from dooers.handlers.send import WorkerSend
from dooers.persistence.base import Persistence
from dooers.protocol.models import (
    Actor,
    AudioPart,
    ContentPart,
    DocumentPart,
    EventType,
    ImagePart,
    Run,
    RunStatus,
    TextPart,
    Thread,
    ThreadEvent,
    User,
    WireC2S_AudioPart,
    WireC2S_ContentPart,
    WireC2S_DocumentPart,
    WireC2S_ImagePart,
    WireC2S_TextPart,
    WireS2C_AudioPart,
    WireS2C_ContentPart,
    WireS2C_DocumentPart,
    WireS2C_ImagePart,
    WireS2C_TextPart,
)
from dooers.registry import ConnectionRegistry
from dooers.repository import Repository
from dooers.server import WorkerServer

__all__ = [
    # Core
    "WorkerConfig",
    "WorkerServer",
    "WorkerContext",
    "WorkerIncoming",
    "WorkerSend",
    "WorkerMemory",
    "ConnectionRegistry",
    "BroadcastManager",
    "Persistence",
    # Dispatch
    "DispatchStream",
    "DispatchError",
    "HandlerError",
    "UploadReferenceError",
    "Handler",
    # Repository
    "Repository",
    # Protocol models
    "User",
    "ContentPart",
    "TextPart",
    "AudioPart",
    "ImagePart",
    "DocumentPart",
    "WireC2S_ContentPart",
    "WireC2S_TextPart",
    "WireC2S_AudioPart",
    "WireC2S_ImagePart",
    "WireC2S_DocumentPart",
    "WireS2C_ContentPart",
    "WireS2C_TextPart",
    "WireS2C_AudioPart",
    "WireS2C_ImagePart",
    "WireS2C_DocumentPart",
    "Thread",
    "ThreadEvent",
    "Run",
    "RunStatus",
    "Actor",
    "EventType",
    # Analytics
    "AnalyticsEvent",
    "AnalyticsEventPayload",
    "AnalyticsBatch",
    "AnalyticsCollector",
    "WorkerAnalytics",
    # Settings
    "SettingsFieldType",
    "SettingsField",
    "SettingsFieldGroup",
    "SettingsSelectOption",
    "SettingsSchema",
    "SettingsBroadcaster",
    "WorkerSettings",
]



from datetime import datetime
from enum import StrEnum
from typing import Any

from pydantic import BaseModel


class AnalyticsEvent(StrEnum):
    """Built-in analytics event types."""

    # Automatic events
    THREAD_CREATED = "thread.created"
    MESSAGE_C2S = "message.c2s"
    MESSAGE_S2C = "message.s2c"
    TOOL_CALLED = "tool.called"
    ERROR_OCCURRED = "error.occurred"

    # Feedback events
    FEEDBACK_LIKE = "feedback.like"
    FEEDBACK_DISLIKE = "feedback.dislike"


class AnalyticsEventPayload(BaseModel):
    """Payload for an analytics event."""

    event: str  # AnalyticsEvent value or custom string
    timestamp: datetime
    worker_id: str
    thread_id: str | None = None
    user_id: str | None = None
    run_id: str | None = None
    event_id: str | None = None
    organization_id: str | None = None
    workspace_id: str | None = None
    data: dict[str, Any] | None = None
    created_at: datetime


class AnalyticsBatch(BaseModel):
    """Batch of analytics events for webhook delivery."""

    batch_id: str
    worker_id: str
    events: list[AnalyticsEventPayload]
    sent_at: datetime



from __future__ import annotations

import uuid
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any

from dooers.protocol.models import ContentPart, Run, Thread, ThreadEvent, User

if TYPE_CHECKING:
    from dooers.persistence.base import Persistence


def _generate_id() -> str:
    return str(uuid.uuid4())


def _now() -> datetime:
    return datetime.now(UTC)


class Repository:
    """Unified CRUD interface for threads, events, runs, and settings."""

    def __init__(self, persistence: Persistence) -> None:
        self._persistence = persistence

    # -- Threads --

    async def list_threads(
        self,
        filter: dict[str, str] | None = None,
        order: dict[str, str] | None = None,
        limit: int = 50,
    ) -> list[Thread]:
        f = filter or {}
        return await self._persistence.list_threads(
            worker_id=f.get("worker_id", ""),
            organization_id=f.get("organization_id", ""),
            workspace_id=f.get("workspace_id", ""),
            user_id=f.get("user_id"),
            cursor=None,
            limit=limit,
            scope=f.get("scope", "member"),
        )

    async def get_thread(self, thread_id: str) -> Thread | None:
        return await self._persistence.get_thread(thread_id)

    async def create_thread(
        self,
        worker_id: str,
        organization_id: str = "",
        workspace_id: str = "",
        user: User | None = None,
        title: str | None = None,
    ) -> Thread:
        now = _now()
        owner = user or User(user_id="")
        thread = Thread(
            id=_generate_id(),
            worker_id=worker_id,
            organization_id=organization_id,
            workspace_id=workspace_id,
            owner=owner,
            users=[owner] if owner.user_id else [],
            title=title,
            created_at=now,
            updated_at=now,
            last_event_at=now,
        )
        await self._persistence.create_thread(thread)
        return thread

    async def update_thread(self, thread_id: str, **updates: Any) -> Thread | None:
        thread = await self._persistence.get_thread(thread_id)
        if not thread:
            return None
        if "title" in updates:
            thread.title = updates["title"]
        thread.updated_at = _now()
        await self._persistence.update_thread(thread)
        return thread

    async def remove_thread(self, thread_id: str) -> None:
        await self._persistence.delete_thread(thread_id)

    # -- Events --

    async def list_events(
        self,
        filter: dict[str, str] | None = None,
        order: dict[str, str] | None = None,
        limit: int = 50,
    ) -> list[ThreadEvent]:
        f = filter or {}
        thread_id = f.get("thread_id", "")
        if not thread_id:
            raise ValueError("filter['thread_id'] is required for list_events")

        event_order = "desc"
        if order and order.get("direction") == "asc":
            event_order = "asc"

        event_filters: dict[str, str] = {}
        for key in ("type", "actor", "user_id", "run_id"):
            if key in f:
                event_filters[key] = f[key]

        return await self._persistence.get_events(
            thread_id,
            limit=limit,
            order=event_order,
            filters=event_filters or None,
        )

    async def get_event(self, event_id: str) -> ThreadEvent | None:
        return await self._persistence.get_event(event_id)

    async def create_event(
        self,
        thread_id: str,
        type: str,
        actor: str,
        content: list[ContentPart] | None = None,
        data: dict | None = None,
        user: User | None = None,
        author: str | None = None,
        run_id: str | None = None,
    ) -> ThreadEvent:
        event = ThreadEvent(
            id=_generate_id(),
            thread_id=thread_id,
            run_id=run_id,
            type=type,
            actor=actor,
            author=author,
            user=user or User(user_id=""),
            content=content,
            data=data,
            created_at=_now(),
        )
        await self._persistence.create_event(event)
        return event

    async def remove_event(self, event_id: str) -> None:
        await self._persistence.delete_event(event_id)

    # -- Runs --

    async def list_runs(
        self,
        filter: dict[str, str] | None = None,
        order: dict[str, str] | None = None,
        limit: int = 50,
    ) -> list[Run]:
        f = filter or {}
        return await self._persistence.list_runs(
            thread_id=f.get("thread_id"),
            worker_id=f.get("worker_id"),
            status=f.get("status"),
            limit=limit,
        )

    async def get_run(self, run_id: str) -> Run | None:
        return await self._persistence.get_run(run_id)

    # -- Settings --

    async def get_settings(self, worker_id: str) -> dict:
        return await self._persistence.get_settings(worker_id)

    async def update_settings(self, worker_id: str, values: dict) -> None:
        await self._persistence.set_settings(worker_id, values)



from dataclasses import dataclass

from dooers.handlers.context import WorkerContext
from dooers.protocol.models import ContentPart


@dataclass
class WorkerIncoming:
    """Represents an incoming message with its complete context.

    Attributes:
      message: Extracted text from content parts
      content: Full content parts from the message
      context: WorkerContext with metadata (thread, org, user info)
    """

    message: str
    content: list[ContentPart]
    context: WorkerContext



from fastapi import FastAPI, WebSocket

from dooers import WorkerConfig, WorkerServer

app = FastAPI()
worker_server = WorkerServer(
    WorkerConfig(
        database_type="sqlite",
        database_name="worker.db",
        assistant_name="Echo Bot",
    )
)


async def echo_agent(incoming, send, memory, analytics, settings):
    yield send.run_start(agent_id="echo")
    yield send.text(f"You said: {incoming.message}")
    yield send.update_thread(title=incoming.message[:60])
    yield send.run_end()


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    await worker_server.handle(websocket, echo_agent)


@app.on_event("shutdown")
async def shutdown():
    await worker_server.close()



import uuid
from dataclasses import dataclass
from typing import Literal


@dataclass
class WorkerEvent:
    send_type: str
    data: dict


class WorkerSend:
    def text(self, text: str, author: str | None = None) -> WorkerEvent:
        return WorkerEvent(
            send_type="text",
            data={"text": text, "author": author},
        )

    def image(
        self,
        url: str,
        mime_type: str | None = None,
        alt: str | None = None,
        author: str | None = None,
    ) -> WorkerEvent:
        return WorkerEvent(
            send_type="image",
            data={"url": url, "mime_type": mime_type, "alt": alt, "author": author},
        )

    def document(
        self,
        url: str,
        filename: str,
        mime_type: str,
        author: str | None = None,
    ) -> WorkerEvent:
        return WorkerEvent(
            send_type="document",
            data={"url": url, "filename": filename, "mime_type": mime_type, "author": author},
        )

    def audio(
        self,
        url: str,
        mime_type: str,
        duration: float | None = None,
        author: str | None = None,
    ) -> WorkerEvent:
        return WorkerEvent(
            send_type="audio",
            data={"url": url, "mime_type": mime_type, "duration": duration, "author": author},
        )

    def tool_call(
        self,
        name: str,
        args: dict,
        display_name: str | None = None,
        id: str | None = None,
    ) -> WorkerEvent:
        return WorkerEvent(
            send_type="tool_call",
            data={
                "id": id or str(uuid.uuid4()),
                "name": name,
                "display_name": display_name,
                "args": args,
            },
        )

    def tool_result(
        self,
        name: str,
        result: dict,
        args: dict | None = None,
        display_name: str | None = None,
        id: str | None = None,
    ) -> WorkerEvent:
        return WorkerEvent(
            send_type="tool_result",
            data={
                "id": id or str(uuid.uuid4()),
                "name": name,
                "display_name": display_name,
                "args": args,
                "result": result,
            },
        )

    def tool_transaction(
        self,
        name: str,
        args: dict,
        result: dict,
        display_name: str | None = None,
        id: str | None = None,
    ) -> WorkerEvent:
        return WorkerEvent(
            send_type="tool_transaction",
            data={
                "id": id or str(uuid.uuid4()),
                "name": name,
                "display_name": display_name,
                "args": args,
                "result": result,
            },
        )

    def run_start(self, agent_id: str | None = None) -> WorkerEvent:
        return WorkerEvent(
            send_type="run_start",
            data={"agent_id": agent_id},
        )

    def run_end(
        self,
        status: Literal["succeeded", "failed"] = "succeeded",
        error: str | None = None,
    ) -> WorkerEvent:
        return WorkerEvent(
            send_type="run_end",
            data={"status": status, "error": error},
        )

    def update_user_event(self, event_id: str, content: list[dict]) -> WorkerEvent:
        return WorkerEvent(
            send_type="event_update",
            data={"event_id": event_id, "content": content},
        )

    def update_thread(self, *, title: str | None = None) -> WorkerEvent:
        return WorkerEvent(
            send_type="thread_update",
            data={"title": title},
        )



from __future__ import annotations

import asyncio
import logging
import time
import uuid
from dataclasses import dataclass

logger = logging.getLogger("workers")

DEFAULT_MAX_SIZE = 25 * 1024 * 1024  # 25MB
DEFAULT_TTL = 300  # 5 minutes
PRUNE_INTERVAL = 60  # seconds


@dataclass
class UploadEntry:
    data: bytes
    mime_type: str
    filename: str
    size_bytes: int
    created_at: float


class UploadStore:
    def __init__(self, max_size: int = DEFAULT_MAX_SIZE, ttl: int = DEFAULT_TTL) -> None:
        self._store: dict[str, UploadEntry] = {}
        self._max_size = max_size
        self._ttl = ttl
        self._prune_task: asyncio.Task | None = None

    async def start(self) -> None:
        if self._prune_task is None:
            self._prune_task = asyncio.create_task(self._prune_loop())

    async def stop(self) -> None:
        if self._prune_task is not None:
            self._prune_task.cancel()
            self._prune_task = None

    def store(self, data: bytes, filename: str, mime_type: str) -> str:
        if len(data) > self._max_size:
            raise ValueError(f"File exceeds maximum size of {self._max_size} bytes")

        ref_id = str(uuid.uuid4())
        self._store[ref_id] = UploadEntry(
            data=data,
            mime_type=mime_type,
            filename=filename,
            size_bytes=len(data),
            created_at=time.time(),
        )
        return ref_id

    def consume(self, ref_id: str) -> UploadEntry | None:
        return self._store.pop(ref_id, None)

    def _prune_expired(self) -> None:
        now = time.time()
        expired = [k for k, v in self._store.items() if now - v.created_at > self._ttl]
        for k in expired:
            del self._store[k]

    async def _prune_loop(self) -> None:
        while True:
            await asyncio.sleep(PRUNE_INTERVAL)
            try:
                self._prune_expired()
            except Exception:
                logger.exception("upload store prune failed")



from dooers.protocol.frames import (
    # Payload types
    AnalyticsSubscribePayload,
    AnalyticsUnsubscribePayload,
    # Analytics C2S frames
    C2S_AnalyticsSubscribe,
    C2S_AnalyticsUnsubscribe,
    # Existing C2S frames
    C2S_Connect,
    C2S_EventCreate,
    C2S_Feedback,
    # Settings C2S frames
    C2S_SettingsPatch,
    C2S_SettingsSubscribe,
    C2S_SettingsUnsubscribe,
    C2S_ThreadDelete,
    C2S_ThreadList,
    C2S_ThreadSubscribe,
    C2S_ThreadUnsubscribe,
    FeedbackAckPayload,
    FeedbackPayload,
    # Existing S2C frames
    S2C_Ack,
    # Analytics S2C frames
    S2C_AnalyticsEvent,
    S2C_EventAppend,
    S2C_FeedbackAck,
    S2C_RunUpsert,
    # Settings S2C frames
    S2C_SettingsPatch,
    S2C_SettingsSnapshot,
    S2C_ThreadDeleted,
    S2C_ThreadListResult,
    S2C_ThreadSnapshot,
    S2C_ThreadUpsert,
    SettingsPatchBroadcastPayload,
    SettingsPatchPayload,
    SettingsSnapshotPayload,
    SettingsSubscribePayload,
    SettingsUnsubscribePayload,
    ThreadDeletedPayload,
    ThreadDeletePayload,
    WSFrame,
)
from dooers.protocol.models import (
    Actor,
    ContentPart,
    DocumentPart,
    EventType,
    ImagePart,
    Run,
    RunStatus,
    TextPart,
    Thread,
    ThreadEvent,
    User,
)
from dooers.protocol.parser import parse_frame, serialize_frame

__all__ = [
    # Models
    "User",
    "ContentPart",
    "TextPart",
    "ImagePart",
    "DocumentPart",
    "Thread",
    "ThreadEvent",
    "Run",
    "RunStatus",
    "Actor",
    "EventType",
    "WSFrame",
    # Existing C2S frames
    "C2S_Connect",
    "C2S_ThreadList",
    "C2S_ThreadDelete",
    "C2S_ThreadSubscribe",
    "C2S_ThreadUnsubscribe",
    "C2S_EventCreate",
    # Analytics C2S frames
    "C2S_AnalyticsSubscribe",
    "C2S_AnalyticsUnsubscribe",
    "C2S_Feedback",
    # Settings C2S frames
    "C2S_SettingsSubscribe",
    "C2S_SettingsUnsubscribe",
    "C2S_SettingsPatch",
    # Existing S2C frames
    "S2C_Ack",
    "S2C_ThreadListResult",
    "S2C_ThreadSnapshot",
    "S2C_EventAppend",
    "S2C_ThreadUpsert",
    "S2C_RunUpsert",
    "S2C_ThreadDeleted",
    "ThreadDeletePayload",
    "ThreadDeletedPayload",
    # Analytics S2C frames
    "S2C_AnalyticsEvent",
    "S2C_FeedbackAck",
    # Settings S2C frames
    "S2C_SettingsSnapshot",
    "S2C_SettingsPatch",
    # Payload types
    "AnalyticsSubscribePayload",
    "AnalyticsUnsubscribePayload",
    "FeedbackPayload",
    "FeedbackAckPayload",
    "SettingsSubscribePayload",
    "SettingsUnsubscribePayload",
    "SettingsPatchPayload",
    "SettingsSnapshotPayload",
    "SettingsPatchBroadcastPayload",
    # Parser
    "parse_frame",
    "serialize_frame",
]



import json
import logging
import ssl as ssl_module
import uuid
from datetime import UTC, datetime
from typing import Any

import asyncpg

from dooers.features.analytics.models import AnalyticsEventPayload
from dooers.protocol.models import (
    Run,
    Thread,
    ThreadEvent,
    User,
    deserialize_s2c_part,
)

logger = logging.getLogger(__name__)


def _build_ssl_context(ssl_config: bool | str) -> ssl_module.SSLContext | str | None:

    if isinstance(ssl_config, str):
        lower = ssl_config.lower().strip()
        if lower in ("false", "0", "no", "off", "", "disable"):
            return None
        if lower in ("true", "1", "yes", "on", "require"):
            ctx = ssl_module.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl_module.CERT_NONE
            return ctx
        return ssl_config

    if ssl_config is True:
        ctx = ssl_module.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl_module.CERT_NONE
        return ctx

    return None


class PostgresPersistence:
    def __init__(
        self,
        *,
        host: str,
        port: int,
        user: str,
        database: str,
        password: str,
        ssl: bool | str = False,
        table_prefix: str = "worker_",
    ):
        self._host = host
        self._port = port
        self._user = user
        self._database = database
        self._password = password
        self._ssl = ssl
        self._prefix = table_prefix
        self._pool: asyncpg.Pool | None = None

    async def connect(self) -> None:
        ssl_param = _build_ssl_context(self._ssl)
        ssl_display = self._ssl if isinstance(self._ssl, str) else ("require" if self._ssl else "disabled")
        logger.info(
            "[workers] connecting to postgresql at %s:%s/%s (user=%s, ssl=%s)",
            self._host,
            self._port,
            self._database,
            self._user,
            ssl_display,
        )
        try:
            self._pool = await asyncpg.create_pool(
                host=self._host,
                port=self._port,
                user=self._user,
                database=self._database,
                password=self._password,
                ssl=ssl_param,
                min_size=1,
                max_size=10,
                timeout=30,
                command_timeout=30,
            )
            logger.info("[workers] successfully connected to postgresql")
        except TimeoutError:
            logger.error(
                "[workers] connection to postgresql timed out. host=%s, port=%s, db=%s, ssl=%s. ",
                self._host,
                self._port,
                self._database,
                ssl_display,
            )
            raise
        except OSError as e:
            logger.error(
                "[workers] cannot reach postgresql at %s:%s - %s. ",
                self._host,
                self._port,
                e,
            )
            raise
        except asyncpg.InvalidPasswordError:
            logger.error(
                "[workers] authentication failed for user '%s' on database '%s'. ",
                self._user,
                self._database,
            )
            raise
        except Exception as e:
            logger.error(
                "[workers] failed to connect to postgresql at %s:%s/%s - %s: %s",
                self._host,
                self._port,
                self._database,
                type(e).__name__,
                e,
            )
            raise

    async def disconnect(self) -> None:
        if self._pool:
            await self._pool.close()

    async def migrate(self) -> None:
        if not self._pool:
            raise RuntimeError("Not connected")

        threads_table = f"{self._prefix}threads"
        events_table = f"{self._prefix}events"
        runs_table = f"{self._prefix}runs"

        async with self._pool.acquire() as conn:
            await conn.execute(f"""
                CREATE TABLE IF NOT EXISTS {threads_table} (
                    id TEXT PRIMARY KEY,
                    worker_id TEXT NOT NULL,
                    organization_id TEXT NOT NULL DEFAULT '',
                    workspace_id TEXT NOT NULL DEFAULT '',
                    owner JSONB NOT NULL DEFAULT '{{}}',
                    users JSONB NOT NULL DEFAULT '[]',
                    title TEXT,
                    created_at TIMESTAMPTZ NOT NULL,
                    updated_at TIMESTAMPTZ NOT NULL,
                    last_event_at TIMESTAMPTZ NOT NULL
                )
            """)

            await conn.execute(f"""
                CREATE TABLE IF NOT EXISTS {events_table} (
                    id TEXT PRIMARY KEY,
                    thread_id TEXT NOT NULL REFERENCES {threads_table}(id),
                    run_id TEXT,
                    type TEXT NOT NULL,
                    actor TEXT NOT NULL,
                    author TEXT,
                    user_id TEXT,
                    user_name TEXT,
                    user_email TEXT,
                    content JSONB,
                    data JSONB,
                    created_at TIMESTAMPTZ NOT NULL
                )
            """)

            await conn.execute(f"""
                CREATE TABLE IF NOT EXISTS {runs_table} (
                    id TEXT PRIMARY KEY,
                    thread_id TEXT NOT NULL REFERENCES {threads_table}(id),
                    agent_id TEXT,
                    status TEXT NOT NULL,
                    started_at TIMESTAMPTZ NOT NULL,
                    ended_at TIMESTAMPTZ,
                    error TEXT
                )
            """)

            await conn.execute(f"""
                ALTER TABLE {events_table} ADD COLUMN IF NOT EXISTS run_id TEXT
            """)
            await conn.execute(f"""
                ALTER TABLE {events_table} ADD COLUMN IF NOT EXISTS user_name TEXT
            """)
            await conn.execute(f"""
                ALTER TABLE {events_table} ADD COLUMN IF NOT EXISTS user_email TEXT
            """)
            await conn.execute(f"""
                ALTER TABLE {events_table} ADD COLUMN IF NOT EXISTS author TEXT
            """)

            await conn.execute(f"""
                ALTER TABLE {threads_table} ADD COLUMN IF NOT EXISTS organization_id TEXT
            """)
            await conn.execute(f"""
                ALTER TABLE {threads_table} ADD COLUMN IF NOT EXISTS workspace_id TEXT
            """)

            await conn.execute(f"""
                ALTER TABLE {threads_table} ADD COLUMN IF NOT EXISTS owner JSONB NOT NULL DEFAULT '{{}}'
            """)
            await conn.execute(f"""
                ALTER TABLE {threads_table} ADD COLUMN IF NOT EXISTS users JSONB NOT NULL DEFAULT '[]'
            """)

            # Migrate legacy user_id column for existing databases
            await conn.execute(f"""
                ALTER TABLE {threads_table} DROP COLUMN IF EXISTS user_id
            """)

            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}threads_users
                    ON {threads_table} USING GIN (users jsonb_path_ops)
            """)

            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}threads_worker_id
                    ON {threads_table}(worker_id)
            """)

            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}threads_organization_id
                    ON {threads_table}(organization_id)
            """)

            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}threads_workspace_id
                    ON {threads_table}(workspace_id)
            """)

            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}events_thread_id
                    ON {events_table}(thread_id)
            """)

            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}events_user_id
                    ON {events_table}(user_id)
            """)

            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}threads_worker_last_event
                    ON {threads_table}(worker_id, last_event_at DESC)
            """)

            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}events_thread_created
                    ON {events_table}(thread_id, created_at)
            """)

            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}runs_thread_id
                    ON {runs_table}(thread_id)
            """)

            settings_table = f"{self._prefix}settings"
            await conn.execute(f"""
                CREATE TABLE IF NOT EXISTS {settings_table} (
                    worker_id TEXT PRIMARY KEY,
                    values JSONB NOT NULL DEFAULT '{{}}',
                    created_at TIMESTAMPTZ DEFAULT NOW(),
                    updated_at TIMESTAMPTZ DEFAULT NOW()
                )
            """)

            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}settings_worker
                    ON {settings_table}(worker_id)
            """)

            analytics_table = f"{self._prefix}analytics_events"
            await conn.execute(f"""
                CREATE TABLE IF NOT EXISTS {analytics_table} (
                    id TEXT PRIMARY KEY,
                    event TEXT NOT NULL,
                    timestamp TIMESTAMPTZ NOT NULL,
                    worker_id TEXT NOT NULL,
                    thread_id TEXT,
                    user_id TEXT,
                    run_id TEXT,
                    event_id TEXT,
                    organization_id TEXT,
                    workspace_id TEXT,
                    data JSONB,
                    created_at TIMESTAMPTZ NOT NULL
                )
            """)
            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}analytics_events_worker_id
                    ON {analytics_table}(worker_id)
            """)
            await conn.execute(f"""
                CREATE INDEX IF NOT EXISTS idx_{self._prefix}analytics_events_timestamp
                    ON {analytics_table}(timestamp)
            """)

    async def create_thread(self, thread: Thread) -> None:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}threads"
        owner_json = json.dumps(thread.owner.model_dump())
        users_json = json.dumps([u.model_dump() for u in thread.users])

        async with self._pool.acquire() as conn:
            await conn.execute(
                f"""
                INSERT INTO {table}
                    (id, worker_id, organization_id, workspace_id, owner, users, title, created_at, updated_at, last_event_at)
                VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10)
                """,
                thread.id,
                thread.worker_id,
                thread.organization_id,
                thread.workspace_id,
                owner_json,
                users_json,
                thread.title,
                thread.created_at,
                thread.updated_at,
                thread.last_event_at,
            )

    async def get_thread(self, thread_id: str) -> Thread | None:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}threads"
        async with self._pool.acquire() as conn:
            row = await conn.fetchrow(
                f"SELECT * FROM {table} WHERE id = $1",
                thread_id,
            )

        if not row:
            return None

        return self._row_to_thread(row)

    def _row_to_thread(self, row: asyncpg.Record) -> Thread:
        owner_data = row["owner"] if row["owner"] else {"user_id": ""}
        if isinstance(owner_data, str):
            owner_data = json.loads(owner_data)

        users_data = row["users"] if row["users"] else []
        if isinstance(users_data, str):
            users_data = json.loads(users_data)

        return Thread(
            id=row["id"],
            worker_id=row["worker_id"],
            organization_id=row["organization_id"] or "",
            workspace_id=row["workspace_id"] or "",
            owner=User(**owner_data),
            users=[User(**u) for u in users_data],
            title=row["title"],
            created_at=row["created_at"],
            updated_at=row["updated_at"],
            last_event_at=row["last_event_at"],
        )

    async def update_thread(self, thread: Thread) -> None:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}threads"
        owner_json = json.dumps(thread.owner.model_dump())
        users_json = json.dumps([u.model_dump() for u in thread.users])

        async with self._pool.acquire() as conn:
            await conn.execute(
                f"""
                UPDATE {table}
                SET organization_id = $1, workspace_id = $2, owner = $3, users = $4,
                    title = $5, updated_at = $6, last_event_at = $7
                WHERE id = $8
                """,
                thread.organization_id,
                thread.workspace_id,
                owner_json,
                users_json,
                thread.title,
                thread.updated_at,
                thread.last_event_at,
                thread.id,
            )

    async def delete_thread(self, thread_id: str) -> None:
        if not self._pool:
            raise RuntimeError("Not connected")

        events_table = f"{self._prefix}events"
        runs_table = f"{self._prefix}runs"
        threads_table = f"{self._prefix}threads"

        async with self._pool.acquire() as conn:
            await conn.execute(f"DELETE FROM {events_table} WHERE thread_id = $1", thread_id)
            await conn.execute(f"DELETE FROM {runs_table} WHERE thread_id = $1", thread_id)
            await conn.execute(f"DELETE FROM {threads_table} WHERE id = $1", thread_id)

    def _build_scope_conditions(
        self,
        scope: str,
        organization_id: str,
        workspace_id: str,
        user_id: str | None,
        user_email: str | None,
        identity_ids: list[str] | None,
        conditions: list[str],
        params: list,
        idx: int,
    ) -> int:
        """Append scope-based WHERE conditions for thread queries. Returns updated idx."""
        if scope == "organization":
            conditions.append(f"organization_id = ${idx}")
            params.append(organization_id)
            idx += 1
        elif scope == "workspace":
            conditions.append(f"organization_id = ${idx}")
            params.append(organization_id)
            idx += 1
            conditions.append(f"workspace_id = ${idx}")
            params.append(workspace_id)
            idx += 1
        elif scope == "member":
            conditions.append(f"organization_id = ${idx}")
            params.append(organization_id)
            idx += 1
            conditions.append(f"workspace_id = ${idx}")
            params.append(workspace_id)
            idx += 1
            if identity_ids:
                # OR'd containment checks use the GIN index on users column
                or_parts: list[str] = []
                for uid in identity_ids:
                    or_parts.append(f"users @> ${idx}::jsonb")
                    params.append(json.dumps([{"user_id": uid}]))
                    idx += 1
                # Also check thread users' stored identity_ids for any match
                or_parts.append(f"""EXISTS (
                    SELECT 1 FROM jsonb_array_elements(users) AS u,
                    jsonb_array_elements_text(COALESCE(u->'identity_ids', '[]'::jsonb)) AS iid
                    WHERE iid = ANY(${idx}::text[])
                )""")
                params.append(identity_ids)
                idx += 1
                conditions.append(f"({' OR '.join(or_parts)})")
            elif user_id:
                conditions.append(f"users @> ${idx}::jsonb")
                params.append(json.dumps([{"user_id": user_id}]))
                idx += 1
            elif user_email:
                conditions.append(f"users @> ${idx}::jsonb")
                params.append(json.dumps([{"user_email": user_email}]))
                idx += 1
        # scope == "admin" — no additional filters
        return idx

    async def count_threads(
        self,
        worker_id: str,
        organization_id: str,
        workspace_id: str,
        user_id: str | None,
        scope: str = "member",
        user_email: str | None = None,
        identity_ids: list[str] | None = None,
    ) -> int:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}threads"
        conditions: list[str] = ["worker_id = $1"]
        params: list[Any] = [worker_id]
        idx = 2

        idx = self._build_scope_conditions(
            scope, organization_id, workspace_id, user_id, user_email, identity_ids,
            conditions, params, idx,
        )

        where = " AND ".join(conditions)
        query = f"SELECT COUNT(*) FROM {table} WHERE {where}"

        async with self._pool.acquire() as conn:
            return await conn.fetchval(query, *params)

    async def list_threads(
        self,
        worker_id: str,
        organization_id: str,
        workspace_id: str,
        user_id: str | None,
        cursor: str | None,
        limit: int,
        scope: str = "member",
        user_email: str | None = None,
        identity_ids: list[str] | None = None,
    ) -> list[Thread]:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}threads"
        conditions: list[str] = ["worker_id = $1"]
        params: list[Any] = [worker_id]
        idx = 2

        idx = self._build_scope_conditions(
            scope, organization_id, workspace_id, user_id, user_email, identity_ids,
            conditions, params, idx,
        )

        if cursor:
            if "|" in cursor:
                cursor_ts, cursor_id = cursor.rsplit("|", 1)
                conditions.append(f"(last_event_at < ${idx} OR (last_event_at = ${idx} AND id < ${idx + 1}))")
                params.append(datetime.fromisoformat(cursor_ts))
                params.append(cursor_id)
                idx += 2
            else:
                conditions.append(f"last_event_at < ${idx}")
                params.append(datetime.fromisoformat(cursor))
                idx += 1

        params.append(limit)
        where = " AND ".join(conditions)

        query = f"""
            SELECT * FROM {table}
            WHERE {where}
            ORDER BY last_event_at DESC, id DESC
            LIMIT ${idx}
        """

        async with self._pool.acquire() as conn:
            rows = await conn.fetch(query, *params)

        return [self._row_to_thread(row) for row in rows]

    async def create_event(self, event: ThreadEvent) -> None:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}events"
        content_json = None
        if event.content:
            content_json = json.dumps([self._serialize_content_part(p) for p in event.content])

        data_json = json.dumps(event.data) if event.data else None

        async with self._pool.acquire() as conn:
            await conn.execute(
                f"""
                INSERT INTO {table} (id, thread_id, run_id, type, actor, author, user_id, user_name, user_email, content, data, created_at)
                VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12)
                """,
                event.id,
                event.thread_id,
                event.run_id,
                event.type,
                event.actor,
                event.author,
                event.user.user_id or None,
                event.user.user_name,
                event.user.user_email,
                content_json,
                data_json,
                event.created_at,
            )

    async def get_events(
        self,
        thread_id: str,
        *,
        after_event_id: str | None = None,
        before_event_id: str | None = None,
        limit: int = 50,
        order: str = "asc",
        filters: dict[str, str] | None = None,
    ) -> list[ThreadEvent]:
        if not self._pool:
            raise RuntimeError("Not connected")

        from dooers.persistence.base import FILTERABLE_FIELDS

        table = f"{self._prefix}events"
        conditions = ["thread_id = $1"]
        params: list[Any] = [thread_id]
        idx = 2

        if after_event_id:
            async with self._pool.acquire() as conn:
                ref_row = await conn.fetchrow(f"SELECT created_at FROM {table} WHERE id = $1", after_event_id)
            if ref_row:
                op = "<" if order == "desc" else ">"
                conditions.append(f"created_at {op} ${idx}")
                params.append(ref_row["created_at"])
                idx += 1

        if before_event_id:
            async with self._pool.acquire() as conn:
                ref_row = await conn.fetchrow(f"SELECT created_at FROM {table} WHERE id = $1", before_event_id)
            if ref_row:
                conditions.append(f"created_at < ${idx}")
                params.append(ref_row["created_at"])
                idx += 1

        if filters:
            for key, value in filters.items():
                if key in FILTERABLE_FIELDS:
                    conditions.append(f"{key} = ${idx}")
                    params.append(value)
                    idx += 1

        direction = "DESC" if order == "desc" else "ASC"
        where = " AND ".join(conditions)
        params.append(limit)

        query = f"SELECT * FROM {table} WHERE {where} ORDER BY created_at {direction} LIMIT ${idx}"

        async with self._pool.acquire() as conn:
            rows = await conn.fetch(query, *params)

        return [self._row_to_event(row) for row in rows]

    def _row_to_event(self, row: asyncpg.Record) -> ThreadEvent:
        content = None
        if row["content"]:
            content_data = json.loads(row["content"])
            content = [self._deserialize_content_part(p) for p in content_data]

        data = json.loads(row["data"]) if row["data"] else None

        return ThreadEvent(
            id=row["id"],
            thread_id=row["thread_id"],
            run_id=row["run_id"],
            type=row["type"],
            actor=row["actor"],
            author=row.get("author"),
            user=User(
                user_id=row["user_id"] or "",
                user_name=row["user_name"],
                user_email=row["user_email"],
            ),
            content=content,
            data=data,
            created_at=row["created_at"],
        )

    async def upsert_thread_participant(self, thread_id: str, user: User) -> None:
        if not self._pool:
            raise RuntimeError("Not connected")

        # Determine match key: prefer user_id, fallback to user_email
        if user.user_id:
            match_field = "user_id"
            match_value = user.user_id
        elif user.user_email:
            match_field = "user_email"
            match_value = user.user_email
        else:
            return  # No identifier to match on

        table = f"{self._prefix}threads"
        user_json = json.dumps(user.model_dump())
        containment_check = json.dumps([{match_field: match_value}])

        async with self._pool.acquire() as conn:
            await conn.execute(
                f"""
                UPDATE {table}
                SET users = (
                    CASE
                        WHEN NOT users @> $2::jsonb
                        THEN users || $3::jsonb
                        ELSE (
                            SELECT jsonb_agg(
                                CASE
                                    WHEN elem->>$5 = $4 THEN $3::jsonb
                                    ELSE elem
                                END
                            )
                            FROM jsonb_array_elements(users) elem
                        )
                    END
                ),
                updated_at = NOW()
                WHERE id = $1
                """,
                thread_id,
                containment_check,
                user_json,
                match_value,
                match_field,
            )

    async def create_run(self, run: Run) -> None:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}runs"
        async with self._pool.acquire() as conn:
            await conn.execute(
                f"""
                INSERT INTO {table} (id, thread_id, agent_id, status, started_at, ended_at, error)
                VALUES ($1, $2, $3, $4, $5, $6, $7)
                """,
                run.id,
                run.thread_id,
                run.agent_id,
                run.status,
                run.started_at,
                run.ended_at,
                run.error,
            )

    async def update_run(self, run: Run) -> None:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}runs"
        async with self._pool.acquire() as conn:
            await conn.execute(
                f"""
                UPDATE {table}
                SET agent_id = $1, status = $2, ended_at = $3, error = $4
                WHERE id = $5
                """,
                run.agent_id,
                run.status,
                run.ended_at,
                run.error,
                run.id,
            )

    async def get_event(self, event_id: str) -> ThreadEvent | None:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}events"
        async with self._pool.acquire() as conn:
            row = await conn.fetchrow(
                f"SELECT * FROM {table} WHERE id = $1",
                event_id,
            )
        if not row:
            return None
        return self._row_to_event(row)

    async def update_event(self, event: ThreadEvent) -> None:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}events"
        content_json = None
        if event.content:
            content_json = json.dumps([self._serialize_content_part(p) for p in event.content])

        async with self._pool.acquire() as conn:
            await conn.execute(
                f"""
                UPDATE {table}
                SET content = $1, run_id = $2, author = $3
                WHERE id = $4
                """,
                content_json,
                event.run_id,
                event.author,
                event.id,
            )

    async def delete_event(self, event_id: str) -> None:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}events"
        async with self._pool.acquire() as conn:
            await conn.execute(f"DELETE FROM {table} WHERE id = $1", event_id)

    async def get_run(self, run_id: str) -> Run | None:
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}runs"
        async with self._pool.acquire() as conn:
            row = await conn.fetchrow(
                f"SELECT * FROM {table} WHERE id = $1",
                run_id,
            )
        if not row:
            return None
        return Run(
            id=row["id"],
            thread_id=row["thread_id"],
            agent_id=row["agent_id"],
            status=row["status"],
            started_at=row["started_at"],
            ended_at=row["ended_at"],
            error=row["error"],
        )

    async def list_runs(
        self,
        thread_id: str | None = None,
        worker_id: str | None = None,
        status: str | None = None,
        limit: int = 50,
    ) -> list[Run]:
        if not self._pool:
            raise RuntimeError("Not connected")

        runs_table = f"{self._prefix}runs"
        threads_table = f"{self._prefix}threads"
        conditions: list[str] = []
        params: list[Any] = []
        idx = 1

        if thread_id:
            conditions.append(f"r.thread_id = ${idx}")
            params.append(thread_id)
            idx += 1
        if worker_id:
            conditions.append(f"t.worker_id = ${idx}")
            params.append(worker_id)
            idx += 1
        if status:
            conditions.append(f"r.status = ${idx}")
            params.append(status)
            idx += 1

        where = f"WHERE {' AND '.join(conditions)}" if conditions else ""
        params.append(limit)

        join = f"JOIN {threads_table} t ON r.thread_id = t.id" if worker_id else ""
        query = f"""
            SELECT r.* FROM {runs_table} r
            {join}
            {where}
            ORDER BY r.started_at DESC
            LIMIT ${idx}
        """

        async with self._pool.acquire() as conn:
            rows = await conn.fetch(query, *params)

        return [
            Run(
                id=row["id"],
                thread_id=row["thread_id"],
                agent_id=row["agent_id"],
                status=row["status"],
                started_at=row["started_at"],
                ended_at=row["ended_at"],
                error=row["error"],
            )
            for row in rows
        ]

    def _serialize_content_part(self, part) -> dict:
        if hasattr(part, "model_dump"):
            return part.model_dump()
        return dict(part)

    def _deserialize_content_part(self, data: dict):
        return deserialize_s2c_part(data)

    async def get_settings(self, worker_id: str) -> dict[str, Any]:
        """Get all stored values for a worker. Returns empty dict if none."""
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}settings"
        async with self._pool.acquire() as conn:
            row = await conn.fetchrow(
                f"SELECT values FROM {table} WHERE worker_id = $1",
                worker_id,
            )

        if not row:
            return {}

        values = row["values"]
        if isinstance(values, str):
            return json.loads(values)
        return values

    async def update_setting(self, worker_id: str, field_id: str, value: Any) -> datetime:
        """Update a single field value. Returns updated_at timestamp."""
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}settings"
        now = datetime.now(UTC)

        current_values = await self.get_settings(worker_id)
        current_values[field_id] = value
        values_json = json.dumps(current_values)

        async with self._pool.acquire() as conn:
            await conn.execute(
                f"""
                INSERT INTO {table} (worker_id, values, created_at, updated_at)
                VALUES ($1, $2, $3, $4)
                ON CONFLICT(worker_id) DO UPDATE SET
                    values = EXCLUDED.values,
                    updated_at = EXCLUDED.updated_at
                """,
                worker_id,
                values_json,
                now,
                now,
            )
        return now

    async def set_settings(self, worker_id: str, values: dict[str, Any]) -> datetime:
        """Replace all settings values. Returns updated_at timestamp."""
        if not self._pool:
            raise RuntimeError("Not connected")

        table = f"{self._prefix}settings"
        now = datetime.now(UTC)
        values_json = json.dumps(values)

        async with self._pool.acquire() as conn:
            await conn.execute(
                f"""
                INSERT INTO {table} (worker_id, values, created_at, updated_at)
                VALUES ($1, $2, $3, $4)
                ON CONFLICT(worker_id) DO UPDATE SET
                    values = EXCLUDED.values,
                    updated_at = EXCLUDED.updated_at
                """,
                worker_id,
                values_json,
                now,
                now,
            )
        return now

    async def insert_analytics_events(self, events: list[AnalyticsEventPayload]) -> None:
        if not self._pool or not events:
            return

        table = f"{self._prefix}analytics_events"
        data_json: str | None
        async with self._pool.acquire() as conn:
            for ev in events:
                data_json = json.dumps(ev.data) if ev.data else None
                await conn.execute(
                    f"""
                    INSERT INTO {table}
                        (id, event, timestamp, worker_id, thread_id, user_id, run_id, event_id,
                         organization_id, workspace_id, data, created_at)
                    VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12)
                    """,
                    str(uuid.uuid4()),
                    ev.event,
                    ev.timestamp,
                    ev.worker_id,
                    ev.thread_id,
                    ev.user_id,
                    ev.run_id,
                    ev.event_id,
                    ev.organization_id,
                    ev.workspace_id,
                    data_json,
                    ev.created_at,
                )



from datetime import datetime
from typing import Annotated, Any, Generic, Literal, TypeVar

from pydantic import BaseModel, Field

from dooers.features.analytics.models import AnalyticsEventPayload
from dooers.features.settings.models import SettingsField, SettingsFieldGroup
from dooers.protocol.models import Run, Thread, ThreadEvent, User, WireC2S_ContentPart

T = TypeVar("T")


class WSFrame(BaseModel, Generic[T]):
    id: str
    type: str
    payload: T


class ConnectPayload(BaseModel):
    worker_id: str
    organization_id: str = ""
    workspace_id: str = ""
    user: User = User(user_id="")
    auth_token: str | None = None
    client: dict | None = None


class ThreadListPayload(BaseModel):
    cursor: str | None = None
    limit: int | None = None


class ThreadSubscribePayload(BaseModel):
    thread_id: str
    after_event_id: str | None = None


class ThreadUnsubscribePayload(BaseModel):
    thread_id: str


class EventCreateEventPayload(BaseModel):
    type: Literal["message"]
    actor: Literal["user"]
    content: list[WireC2S_ContentPart]
    data: dict | None = None


class EventCreatePayload(BaseModel):
    thread_id: str | None = None
    client_event_id: str | None = None
    event: EventCreateEventPayload


class C2S_Connect(BaseModel):
    id: str
    type: Literal["connect"]
    payload: ConnectPayload


class C2S_ThreadList(BaseModel):
    id: str
    type: Literal["thread.list"]
    payload: ThreadListPayload


class C2S_ThreadSubscribe(BaseModel):
    id: str
    type: Literal["thread.subscribe"]
    payload: ThreadSubscribePayload


class C2S_ThreadUnsubscribe(BaseModel):
    id: str
    type: Literal["thread.unsubscribe"]
    payload: ThreadUnsubscribePayload


class ThreadDeletePayload(BaseModel):
    thread_id: str


class C2S_ThreadDelete(BaseModel):
    id: str
    type: Literal["thread.delete"]
    payload: ThreadDeletePayload


class C2S_EventCreate(BaseModel):
    id: str
    type: Literal["event.create"]
    payload: EventCreatePayload


class AnalyticsSubscribePayload(BaseModel):
    worker_id: str


class AnalyticsUnsubscribePayload(BaseModel):
    worker_id: str


class FeedbackPayload(BaseModel):
    target_type: Literal["event", "run", "thread"]
    target_id: str
    feedback: Literal["like", "dislike"]
    reason: str | None = None
    classification: str | None = None


class C2S_AnalyticsSubscribe(BaseModel):
    id: str
    type: Literal["analytics.subscribe"]
    payload: AnalyticsSubscribePayload


class C2S_AnalyticsUnsubscribe(BaseModel):
    id: str
    type: Literal["analytics.unsubscribe"]
    payload: AnalyticsUnsubscribePayload


class C2S_Feedback(BaseModel):
    id: str
    type: Literal["feedback"]
    payload: FeedbackPayload


class EventListPayload(BaseModel):
    thread_id: str
    before_event_id: str | None = None
    limit: int = 50


class C2S_EventList(BaseModel):
    id: str
    type: Literal["event.list"] = "event.list"
    payload: EventListPayload


class SettingsSubscribePayload(BaseModel):
    worker_id: str


class SettingsUnsubscribePayload(BaseModel):
    worker_id: str


class SettingsPatchPayload(BaseModel):
    field_id: str
    value: Any


class C2S_SettingsSubscribe(BaseModel):
    id: str
    type: Literal["settings.subscribe"]
    payload: SettingsSubscribePayload


class C2S_SettingsUnsubscribe(BaseModel):
    id: str
    type: Literal["settings.unsubscribe"]
    payload: SettingsUnsubscribePayload


class C2S_SettingsPatch(BaseModel):
    id: str
    type: Literal["settings.patch"]
    payload: SettingsPatchPayload


ClientToServer = Annotated[
    C2S_Connect
    | C2S_ThreadList
    | C2S_ThreadSubscribe
    | C2S_ThreadUnsubscribe
    | C2S_ThreadDelete
    | C2S_EventCreate
    | C2S_EventList
    | C2S_AnalyticsSubscribe
    | C2S_AnalyticsUnsubscribe
    | C2S_Feedback
    | C2S_SettingsSubscribe
    | C2S_SettingsUnsubscribe
    | C2S_SettingsPatch,
    Field(discriminator="type"),
]


class AckPayload(BaseModel):
    ack_id: str
    ok: bool
    error: dict | None = None


class ThreadListResultPayload(BaseModel):
    threads: list[Thread]
    cursor: str | None = None
    total_count: int = 0


class ThreadSnapshotPayload(BaseModel):
    thread: Thread
    events: list[ThreadEvent]
    runs: list[Run] | None = None


class EventAppendPayload(BaseModel):
    thread_id: str
    events: list[ThreadEvent]


class ThreadUpsertPayload(BaseModel):
    thread: Thread


class RunUpsertPayload(BaseModel):
    run: Run


class S2C_Ack(BaseModel):
    id: str
    type: Literal["ack"] = "ack"
    payload: AckPayload


class S2C_ThreadListResult(BaseModel):
    id: str
    type: Literal["thread.list.result"] = "thread.list.result"
    payload: ThreadListResultPayload


class S2C_ThreadSnapshot(BaseModel):
    id: str
    type: Literal["thread.snapshot"] = "thread.snapshot"
    payload: ThreadSnapshotPayload


class S2C_EventAppend(BaseModel):
    id: str
    type: Literal["event.append"] = "event.append"
    payload: EventAppendPayload


class S2C_ThreadUpsert(BaseModel):
    id: str
    type: Literal["thread.upsert"] = "thread.upsert"
    payload: ThreadUpsertPayload


class EventListResultPayload(BaseModel):
    thread_id: str
    events: list[ThreadEvent]
    cursor: str | None = None
    has_more: bool = False


class S2C_EventListResult(BaseModel):
    id: str
    type: Literal["event.list.result"] = "event.list.result"
    payload: EventListResultPayload


class ThreadDeletedPayload(BaseModel):
    thread_id: str


class S2C_ThreadDeleted(BaseModel):
    id: str
    type: Literal["thread.deleted"] = "thread.deleted"
    payload: ThreadDeletedPayload


class S2C_RunUpsert(BaseModel):
    id: str
    type: Literal["run.upsert"] = "run.upsert"
    payload: RunUpsertPayload


class S2C_AnalyticsEvent(BaseModel):
    id: str
    type: Literal["analytics.event"] = "analytics.event"
    payload: AnalyticsEventPayload


class FeedbackAckPayload(BaseModel):
    target_type: str
    target_id: str
    feedback: str
    ok: bool


class S2C_FeedbackAck(BaseModel):
    id: str
    type: Literal["feedback.ack"] = "feedback.ack"
    payload: FeedbackAckPayload


class SettingsSnapshotPayload(BaseModel):
    worker_id: str
    fields: list[SettingsField | SettingsFieldGroup]
    updated_at: datetime


class S2C_SettingsSnapshot(BaseModel):
    id: str
    type: Literal["settings.snapshot"] = "settings.snapshot"
    payload: SettingsSnapshotPayload


class SettingsPatchBroadcastPayload(BaseModel):
    worker_id: str
    field_id: str
    value: Any
    updated_at: datetime


class S2C_SettingsPatch(BaseModel):
    id: str
    type: Literal["settings.patch"] = "settings.patch"
    payload: SettingsPatchBroadcastPayload


ServerToClient = (
    S2C_Ack
    | S2C_ThreadListResult
    | S2C_ThreadSnapshot
    | S2C_EventAppend
    | S2C_EventListResult
    | S2C_ThreadUpsert
    | S2C_ThreadDeleted
    | S2C_RunUpsert
    | S2C_AnalyticsEvent
    | S2C_FeedbackAck
    | S2C_SettingsSnapshot
    | S2C_SettingsPatch
)
